from tempfile import NamedTemporaryFile

import numpy as np
import pymzml
import pytest
from scipy.signal import find_peaks
from scipy.stats import kstest, normaltest

import smiter
from smiter.fragmentation_functions import (
    AbstractFragmentor,
    NucleosideFragmentor,
    PeptideFragmentor,
)
from smiter.noise_functions import GaussNoiseInjector, UniformNoiseInjector
from smiter.synthetic_mzml import (
    generate_interval_tree,
    generate_molecule_isotopologue_lib,
    generate_scans,
    write_mzml,
)


class TestFragmentor(AbstractFragmentor):
    def __init__(self):
        print("Fragmentor goes chop chop chop chop chop ...")

    def fragment(self, mol):
        return np.array([(200, 1e5)])


fragmentor = TestFragmentor()
noise_injector = GaussNoiseInjector(variance=0.05)


def test_write_mzml():
    """Write a mzML without spectra readable by pymzML."""
    file = NamedTemporaryFile("wb")
    peak_props = {}
    mzml_params = {
        "gradient_length": 0,
    }
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 0


def test_write_empty_mzml():
    """Write a mzML without spectra readable by pymzML."""
    file = NamedTemporaryFile("wb")
    peak_props = {}
    mzml_params = {
        "gradient_length": 5,
    }
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 167


def test_write_inosine_flat_mzml():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "inosine": {
            "chemical_formula": "+C(10)H(12)N(4)O(5)",
            "trivial_name": "inosine",
            "charge": 2,
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": None,
            "peak_params": {},
        }
    }
    mzml_params = {
        "gradient_length": 30,
    }
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1000


def test_write_inosine_gauss_mzml():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "inosine": {
            "chemical_formula": "+C(10)H(12)N(4)O(5)",
            "trivial_name": "inosine",
            "charge": 2,
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width
        }
    }
    mzml_params = {
        "gradient_length": 30,
    }
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1000
    intensities = []
    for spec in reader:
        if spec.ms_level == 1:
            if len(spec.i) > 0:
                intensities.append(spec.i[0])
    _, p = normaltest(intensities)
    print(intensities)
    # assert p < 5e-4


def test_write_mzml_get_TIC():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "inosine": {
            "chemical_formula": "+C(10)H(12)N(4)O(5)",
            "trivial_name": "inosine",
            "charge": 2,
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": None,
            "peak_params": {},
        }
    }
    mzml_params = {
        "gradient_length": 30,
    }
    noise_injector = GaussNoiseInjector(variance=0.0)
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1000
    reader = pymzml.run.Reader(mzml_path)
    tics = []
    for spec in reader:
        if spec.ms_level == 1:
            tics.append(sum(spec.i))
    tic = reader["TIC"]
    assert tic.peaks()[:, 1] == pytest.approx(np.array(tics, dtype="float32"))


def test_write_inosine_gamma_mzml():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "inosine": {
            "chemical_formula": "+C(10)H(12)N(4)O(5)",
            "trivial_name": "inosine",
            "charge": 2,
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gamma",
            "peak_scaling_factor": 1e6,
            "peak_params": {"a": 3, "scale": 20},  # 10% of peak width,
        }
    }
    mzml_params = {
        "gradient_length": 30,
    }
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1000
    intensities = []
    for spec in reader:
        if spec.ms_level == 1:
            if len(spec.i) > 0:
                intensities.append(spec.i[0])
    ## what is a reasonable p-value cutoff here?
    # assert t.pvalue < 1e-100


def test_write_inosine_adenosine_gauss_mzml():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "inosine": {
            "chemical_formula": "+C(10)H(12)N(4)O(5)",
            "charge": 1,
            "trivial_name": "inosine",
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width,
            "peak_scaling_factor": 1e3,
        },
        "+C(10)H(13)N(5)O(4)": {
            "chemical_formula": "+C(10)H(13)N(5)O(4)",
            "trivial_name": "adenosine",
            "charge": 1,
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width,
        },
    }
    mzml_params = {"gradient_length": 30, "min_intensity": 0}
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1000

    ino_rt = []
    ino_intensities = []
    adeno_rt = []
    adeno_intensities = []

    ino_mono = 269.0880
    adeno_mono = 268.1040

    for spec in reader:
        if spec.ms_level == 1:
            ino_i = spec.i[abs(spec.mz - ino_mono) < 0.001]
            adeno_i = spec.i[abs(spec.mz - adeno_mono) < 0.001]
            if len(adeno_i) > 0:
                adeno_intensities.append(adeno_i[0])
                adeno_rt.append(spec.scan_time[0])
            if len(ino_i) > 0:
                ino_intensities.append(ino_i[0])
                ino_rt.append(spec.scan_time[0])

    _, p = normaltest(ino_intensities)
    # assert p < 5e-4
    _, p = normaltest(adeno_intensities)
    # assert p < 5e-4


def test_write_inosine_adenosine_gauss_shift_mzml():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "inosine": {
            "chemical_formula": "+C(10)H(12)N(4)O(5)",
            "charge": 1,
            "trivial_name": "inosine",
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width,
            "peak_scaling_factor": 1e6,
        },
        "+C(10)H(13)N(5)O(4)": {
            "chemical_formula": "+C(10)H(13)N(5)O(4)",
            "trivial_name": "adenosine",
            "charge": 1,
            "scan_start_time": 15,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width,
            "peak_scaling_factor": 1e6,
        },
    }
    mzml_params = {
        "gradient_length": 45,
    }
    noise_injector = GaussNoiseInjector(variance=0.0)
    # noise_injector = UniformformNoiseInjector()
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1500

    ino_rt = []
    ino_intensities = []
    adeno_rt = []
    adeno_intensities = []

    ino_mono = 269.0880
    adeno_mono = 268.1040

    for spec in reader:
        if spec.ms_level == 1:
            ino_i = spec.i[abs(spec.mz - ino_mono) < 0.001]
            adeno_i = spec.i[abs(spec.mz - adeno_mono) < 0.001]
            if len(adeno_i) > 0:
                adeno_intensities.append(adeno_i[0])
                adeno_rt.append(spec.scan_time[0])
            if len(ino_i) > 0:
                ino_intensities.append(ino_i[0])
                ino_rt.append(spec.scan_time[0])
    # assert rt max diff is about 15
    m_i = np.argmax(ino_intensities)
    m_a = np.argmax(adeno_intensities)
    mean_i_rt = ino_rt[m_i]
    mean_a_rt = adeno_rt[m_a]
    assert 14 < (mean_a_rt - mean_i_rt) < 16


def test_write_inosine_proper_fragments_mzml():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "inosine": {
            "chemical_formula": "+C(10)H(12)N(4)O(5)",
            "charge": 1,
            "trivial_name": "inosine",
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width,
            "peak_scaling_factor": 1e6,
        },
    }
    mzml_params = {
        "gradient_length": 30,
    }

    nucl_fragmentor = NucleosideFragmentor()
    mzml_path = write_mzml(
        file, peak_props, nucl_fragmentor, noise_injector, mzml_params
    )
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1000

    ino_rt = []
    ino_intensities = []
    ino_fragments = []

    ino_mono = 269.0880

    for spec in reader:
        if spec.ms_level == 1:
            ino_i = spec.i[abs(spec.mz - ino_mono) < 0.001]
            if len(ino_i) > 0:
                ino_intensities.append(ino_i[0])
                ino_rt.append(spec.scan_time[0])

    _, p = normaltest(ino_intensities)
    expected_frags = np.array([137.0457872316])

    for frag_list in ino_fragments:
        assert len(frag_list) == len(expected_frags)
        sorted_frags = np.sort(frag_list)
        assert abs(sorted_frags - expected_frags) < 0.001


@pytest.mark.slow()
def test_write_peptide_gauss_mzml():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "ELVISLIVES": {
            "chemical_formula": "ELVISLIVES",
            "trivial_name": "ELVISLIVES",
            "charge": 2,
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width
        },
        "ELVISLIVSE": {
            "charge": 2,
            "trivial_name": "ELVISLIVSE",
            "chemical_formula": "ELVISLIVSE",
            "scan_start_time": 15,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width,
        },
    }
    mzml_params = {
        "gradient_length": 45,
    }
    fragmentor = PeptideFragmentor()
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    assert reader.get_spectrum_count() == 1500
    intensities = []
    for spec in reader:
        if spec.ms_level == 1:
            if len(spec.i) > 0:
                intensities.append(sum(spec.i))
    _, p = normaltest(intensities)
    # assert p < 5e-4


def test_write_2_mols_same_cc():
    file = NamedTemporaryFile("wb")
    peak_props = {
        "uridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 0.5 * 1e6,
        },
        "pseudouridine": {
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "pseudouridine",
            "charge": 2,
            "scan_start_time": 15,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 1e6,
        },
    }
    mzml_params = {
        "gradient_length": 45,
        "min_intensity": 10,
    }
    mzml_params = {
        "gradient_length": 45,
    }
    fragmentor = NucleosideFragmentor()
    noise_injector = GaussNoiseInjector(variance=0.0)
    mzml_path = write_mzml(file, peak_props, fragmentor, noise_injector, mzml_params)
    reader = pymzml.run.Reader(mzml_path)
    # assert reader.get_spectrum_count() == 1499
    intensities = []
    for spec in reader:
        if spec.ms_level == 1:
            if len(spec.i) > 0:
                intensities.append(sum(spec.i))
    peaks, _ = find_peaks(intensities)
    assert len(peaks) == 2


def test_rescale_intensity():
    i = 100
    rt = 15
    molecule = ""
    peak_props = {
        "uridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 0.5,
        }
    }
    iso_lib = {}

    i_rescaled = smiter.synthetic_mzml.rescale_intensity(
        i, rt, "uridine", peak_props, iso_lib
    )
    # should be max_i @ half peak width, with half peak width = 15 and max_i = 50 (100 * 0.5)
    assert round(i_rescaled, 2) == 50


def test_generate_interval_tree():
    peak_props = {
        "uridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 0,
            "peak_width": 5,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 1,
        },
        "pseudouridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 4,
            "peak_width": 5,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 1,
        },
    }
    tree = generate_interval_tree(peak_props)

    # only uridine
    tree_3 = tree[3]
    assert len(tree_3) == 1
    for entry in tree_3:
        assert entry.data == "uridine"
        assert entry.length() == 5

    # only pseudouridine
    tree_8 = tree[8]
    assert len(tree_3) == 1
    for entry in tree_8:
        assert entry.data == "pseudouridine"
        assert entry.length() == 5

    # uridine and pseudouridine
    tree_4 = tree[4]
    assert len(tree_4) == 2
    items = ["pseudouridine", "uridine"]
    for entry in tree_4:
        items.remove(entry.data)
        assert entry.length() == 5
    assert len(items) == 0


def test_generate_scans_simple():
    # generate 1 MS1 and 1 MS2 scan
    # peak width is as long as gradient and two times ms_rt_diff
    peak_props = {
        "uridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 0,
            "peak_width": 0.06,  # peak as long as ms_rt_diff and gradient length
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 1,
        }
    }
    trivial_names = {val["chemical_formula"]: key for key, val in peak_props.items()}
    scans, mol_scan_dict = generate_scans(
        generate_molecule_isotopologue_lib(peak_props, [1, 2, 3], trivial_names),
        peak_props,
        generate_interval_tree(peak_props),
        NucleosideFragmentor(),
        UniformNoiseInjector(),
        {"gradient_length": 0.06, "min_intensity": 0, "isolation_window_width": 0.2},
    )
    assert len(scans) == 1  # one MS1 with related MS2 scans
    assert len(scans[0]) == 2  # first element is a MS1 scan and a list of MS2 scans
    assert len(scans[0][1]) == 1  # list of MS2 scans has one scan


def test_generate_molecule_isotopologue_lib():
    peak_props = {
        "uridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 0,
            "peak_width": 0.06,  # peak as long as ms_rt_diff and gradient length
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 1,
        }
    }
    charges = [1, 2, 3]
    trivial_names = {"+C(9)H(11)N(2)O(6)": "uridine"}
    lib = generate_molecule_isotopologue_lib(peak_props, charges, trivial_names)
    assert list(lib.keys()) == ["uridine"]
    assert np.isclose(
        lib["uridine"]["mz"],
        [
            122.53813200786999,
            123.03962222482085,
            123.54123939611428,
            124.04272136528171,
            124.5442376065953,
            125.04387103562883,
            125.54630817011011,
        ],
    ).all()
    assert np.isclose(
        lib["uridine"]["i"],
        [
            1.0,
            0.10819885003176334,
            0.01762911869579814,
            0.0013546885053562444,
            6.246133599652487e-05,
            4.563511253647458e-07,
            4.800737885333426e-10,
        ],
    ).all()


def test_write_mzml_one_spec():
    tempfile = NamedTemporaryFile("wb")
    peak_props = {
        "uridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 0,
            "peak_width": 0.03,  # peak as long as ms_rt_diff and gradient length
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 1,
        }
    }
    fragmentor = NucleosideFragmentor()
    noise = UniformNoiseInjector()
    mzml_params = {"gradient_length": 0.03, "ms_rt_diff": 0.03}

    write_mzml(tempfile, peak_props, fragmentor, noise, mzml_params)
    reader = pymzml.run.Reader(tempfile.name)
    assert reader.get_spectrum_count() == 1


def test_dynacmic_exclusion():
    tempfile = NamedTemporaryFile("wb")
    peak_props = {
        "uridine": {
            "charge": 2,
            "chemical_formula": "+C(9)H(11)N(2)O(6)",
            "trivial_name": "uridine",
            "scan_start_time": 0,
            "peak_width": 5,  # peak as long as ms_rt_diff and gradient length
            "peak_function": "gauss",
            "peak_params": {"sigma": 1},  # 10% of peak width,
            "peak_scaling_factor": 1e5,
        }
    }
    trivial_names = {"+C(9)H(11)N(2)O(6)": "uridine"}
    # dynamic_exclusion is bigger than gradient length, so we expect only one MS2 fragment spectrum
    default_mzml_params = {
        "gradient_length": 5,
        "min_intensity": 100,
        "isolation_window_width": 0.5,
        "ion_target": 3e6,
        "ms_rt_diff": 0.03,
        "dynamic_exclusion": 30,  # in seconds
    }
    scans, mol_scan_dict = generate_scans(
        generate_molecule_isotopologue_lib(peak_props, [1, 2, 3], trivial_names),
        peak_props,
        generate_interval_tree(peak_props),
        NucleosideFragmentor(),
        UniformNoiseInjector(),
        default_mzml_params,
    )
    number_fragment_specs = 0
    for ms1, ms2_list in scans:
        for ms2 in ms2_list:
            if len(ms2["mz"]) > 0:
                number_fragment_specs += 1
    # fails when dynamic_exclusion is set to 0
    # since there will be 15 fragments specs
    assert number_fragment_specs == 1

    # breakpoint()



"""Core functionality."""
import csv
from io import TextIOWrapper
from tempfile import _TemporaryFileWrapper

from loguru import logger

from smiter.params.default_params import default_mzml_params, default_peak_properties

PROTON = 1.00727646677


def calc_mz(mass: float, charge: int):
    """Calculate m/z.

    Args:
        mass (TYPE): Description
        charge (TYPE): Description
    """
    mass = float(mass)
    charge = int(charge)
    calc_mz = (mass + (charge * PROTON)) / charge
    return calc_mz


def check_mzml_params(mzml_params: dict) -> dict:
    """Summary.

    Args:
        mzml_params (dict): Description

    Returns:
        dict: Description

    Raises:
        Exception: Description
    """
    logger.info("Checking mzML params")
    for default_param, default_value in default_mzml_params.items():
        # param not set and default param required
        if (mzml_params.get(default_param, None) is None) and (default_value is None):
            raise Exception(f"mzml parameter {default_param} is required by not set!")
        elif mzml_params.get(default_param, None) is None:
            mzml_params[default_param] = default_value
    return mzml_params


def check_peak_properties(peak_properties: dict) -> dict:
    """Summary.

    Args:
        peak_properties (dict): Description

    Returns:
        dict: Description

    Raises:
        Exception: Description
    """
    logger.info("Checking peak properties")
    for mol, properties in peak_properties.items():
        for default_param, default_value in default_peak_properties.items():
            if (properties.get(default_param, None) is None) and (
                default_value is None
            ):
                raise Exception(
                    f"mzml parameter {default_param} is required by not set!"
                )
            elif properties.get(default_param, None) is None:
                properties[default_param] = default_value
    return peak_properties


def csv_to_peak_properties(csv_file):
    logger.info(f"Read peak properties from {csv_file}")
    peak_properties = {}
    with open(csv_file) as fin:
        reader = csv.DictReader(fin)
        for line_dict in reader:
            cc = line_dict["chemical_formula"]
            tn = line_dict["trivial_name"]
            peak_properties[tn] = {
                "trivial_name": line_dict["trivial_name"],
                "chemical_formula": cc,
                "charge": int(line_dict.get("charge", 2)),
                "scan_start_time": float(line_dict["scan_start_time"]),
                # currently only gaussian peaks from csv
                "peak_function": "gauss_tail",
                # "peak_function": "gauss",
                "peak_params": {"sigma": float(line_dict.get("sigma", 2))},
                "peak_scaling_factor": float(line_dict["peak_scaling_factor"]),
                "peak_width": float(line_dict.get("peak_width", 30)),
            }
    return peak_properties


def peak_properties_to_csv(peak_properties, csv_file):
    logger.info(f"Write peak properties to {csv_file}")
    if not isinstance(csv_file, TextIOWrapper):
        csv_file = open(csv_file, "w")
    csv_filename = csv_file.name
    fieldnames = [
        "chemical_formula",
        "trivial_name",
        "charge",
        "scan_start_time",
        "peak_width",
        "peak_scaling_factor",
        "peak_function",
        "peak_params",
    ]
    writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
    writer.writeheader()
    for trivial_name, attribs in peak_properties.items():
        line = {
            "chemical_formula": peak_properties[trivial_name].get(
                "chemical_formula", ""
            ),
            # "trivial_name": peak_properties[cc].get("trivial_name", ""),
            "trivial_name": trivial_name,
            "charge": peak_properties[trivial_name].get("charge", 2),
            "scan_start_time": peak_properties[trivial_name]["scan_start_time"],
            "peak_function": peak_properties[trivial_name]["peak_function"],
            "peak_params": ",".join(
                [
                    f"{key}={val}"
                    for key, val in peak_properties[trivial_name]["peak_params"].items()
                ]
            ),
            "peak_scaling_factor": peak_properties[trivial_name].get(
                "peak_scaling_factor", 1e3
            ),
            "peak_width": peak_properties[trivial_name]["peak_width"],
        }
        writer.writerow(line)
    csv_file.close()
    return csv_filename



"""Summary."""
import os
import numpy as np
import pytest

import smiter
from smiter.fragmentation_functions import (
    NucleosideFragmentor,
    PeptideFragmentor,
    LipidFragmentor,
)


def test_fragment_peptide():
    """Summary."""
    fragger = PeptideFragmentor(charges=1, ions=["y"])
    peaks = fragger.fragment("L")
    expected_mzs = np.array([132.101])
    assert ((peaks[:, 0] - expected_mzs) < 0.001).all()


def test_fragment_nucleotide():
    """Summary."""
    fragger = NucleosideFragmentor()
    peaks = fragger.fragment("adenosine")
    expected_mzs = np.array([119.03522254717, 136.0617716478])
    assert ((peaks[:, 0] - expected_mzs) < 0.001).all()


@pytest.mark.skipif("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true", reason="Skipping this test on Travis CI.")
def test_fragment_lipid():
    test_lipid_file = "test_lipids.txt"
    test_lipid_file = os.path.abspath(test_lipid_file)
    with open(test_lipid_file, "wt") as fout:
        fout.write("PC 18:0/12:0\n")
        fout.write("PE 18:3;1-16:2")
    fragger = LipidFragmentor(test_lipid_file)
    masses = fragger.fragment("PC 18:0/12:0")
    assert np.allclose(masses, np.array([184.0733]))



#!/usr/bin/env python
import errno
import fileinput
import sys


def main():
    try:
        for i, line in enumerate(fileinput.input()):
            if i == 0:
                sys.stdout.write(line)
                continue
            line = line.split(",")[2:]
            sys.stdout.write(",".join(line))
        sys.stdout.flush()
    except IOError as e:
        if e.errno == errno.EPIPE:
            sys.exit(1)


if __name__ == "__main__":
    main()



# None means required, other is default value
# non existent params are optional
default_peak_properties = {
    # trivial_name
    "charge": 2,
    "scan_start_time": None,  # required
    "peak_width": None,  # required
    # "peak_function": "gauss",
}

default_mzml_params = {
    "gradient_length": None,
    "min_intensity": 100,
    "max_intensity": 1e10,  # what would be reasonable?
    "isolation_window_width": 0.5,
    "ion_target": 3e6,
    "ms_rt_diff": 0.03,
    "dynamic_exclusion": 30,  # in seconds
    "max_ms2_spectra": 10,
    "mz_lower_limit": 100,
    "mz_upper_limit": 1600,
}



"""Top-level package for SMITER."""

__author__ = """Manuel Kösters"""
__email__ = "manuel.koesters@dcb.unibe.ch"
__version__ = "0.1.0"

import os
import sys
import tempfile

from loguru import logger

import smiter.synthetic_mzml

config = {
    "handlers": [
        {"sink": sys.stdout, "level": "INFO"},
        {
            "sink": os.path.join(tempfile.gettempdir(), "smiter.log"),
            "compression": "gz",
            "rotation": "100 MB",
            "level": "DEBUG",
            "backtrace": True,
        },
    ],
}
logger.configure(**config)



"""Main module."""
import io
import pathlib
import time
import warnings
from pprint import pformat
from typing import Callable, Dict, List, Tuple, Union
from collections import Counter

import numpy as np
import pyqms
import scipy as sci
from intervaltree import IntervalTree
from loguru import logger
from psims.mzml import MzMLWriter
from tqdm import tqdm

import smiter
from smiter.fragmentation_functions import AbstractFragmentor
from smiter.lib import (
    calc_mz,
    check_mzml_params,
    check_peak_properties,
    peak_properties_to_csv,
)
from smiter.noise_functions import AbstractNoiseInjector
from smiter.peak_distribution import distributions

warnings.filterwarnings("ignore")


class Scan(dict):
    """Summary."""

    def __init__(self, data: dict = None):
        """Summary.

        Args:
            dict (TYPE): Description
        """
        if data is not None:
            self.update(data)

    @property
    def mz(self):
        """Summary."""
        v = self.get("mz", None)
        return v

    @mz.setter
    def mz(self, mz):
        """Summary."""
        self["mz"] = mz

    @property
    def i(self):
        """Summary."""
        v = self.get("i", None)
        return v

    @i.setter
    def i(self, i):
        """Summary."""
        self["i"] = i

    @property
    def id(self):
        """Summary."""
        v = self.get("id", None)
        return v

    @property
    def precursor_mz(self):
        """Summary."""
        v = self.get("precursor_mz", None)
        return v

    @property
    def precursor_i(self):
        """Summary."""
        v = self.get("precursor_i", None)
        return v

    @property
    def precursor_charge(self):
        """Summary."""
        v = self.get("precursor_charge", None)
        return v

    # @property
    # def precursor_scan_id(self):
    #     """Summary."""
    #     v = self.get("precursor_scan_id", None)
    #     return v

    @property
    def retention_time(self):
        """Summary."""
        v = self.get("rt", None)
        return v

    @property
    def ms_level(self):
        """Summary.

        Returns:
            TYPE: Description
        """
        v = self.get("ms_level", None)
        return v


def generate_interval_tree(peak_properties):
    """Conctruct an interval tree containing the elution windows of the analytes.

    Args:
        peak_properties (dict): Description

    Returns:
        IntervalTree: Description
    """
    tree = IntervalTree()
    for key, data in peak_properties.items():
        start = data["scan_start_time"]
        end = start + data["peak_width"]
        tree[start:end] = key
    return tree


# @profile
def write_mzml(
    file: Union[str, io.TextIOWrapper],
    peak_properties: Dict[str, dict],
    fragmentor: AbstractFragmentor,
    noise_injector: AbstractNoiseInjector,
    mzml_params: Dict[str, Union[int, float, str]],
) -> str:
    """Write mzML file with chromatographic peaks and fragment spectra for the given molecules.

    Args:
        file (Union[str, io.TextIOWrapper]): Description
        molecules (List[str]): Description
        fragmentation_function (Callable[[str], List[Tuple[float, float]]], optional): Description
        peak_properties (Dict[str, dict], optional): Description
    """
    # check params and raise Exception(s) if necessary
    logger.info("Start generating mzML")
    mzml_params = check_mzml_params(mzml_params)
    peak_properties = check_peak_properties(peak_properties)

    interval_tree = generate_interval_tree(peak_properties)

    filename = file if isinstance(file, str) else file.name
    scans = []

    trivial_names = {}
    charges = set()
    for key, val in peak_properties.items():
        trivial_names[val["chemical_formula"]] = key
        charges.add(val["charge"])
    # trivial_names = {
    # val["chemical_formula"]: key for key, val in peak_properties.items()
    # }
    # dicts are sorted, language specification since python 3.7+

    isotopologue_lib = generate_molecule_isotopologue_lib(
        peak_properties, trivial_names=trivial_names, charges=charges
    )
    scans, scan_dict = generate_scans(
        isotopologue_lib,
        peak_properties,
        interval_tree,
        fragmentor,
        noise_injector,
        mzml_params,
    )
    logger.info("Delete interval tree")
    del interval_tree
    write_scans(file, scans)
    if not isinstance(file, str):
        file_path = file.name
    else:
        file_path = file
    path = pathlib.Path(file_path)
    summary_path = path.parent.resolve() / "molecule_summary.csv"
    peak_properties_to_csv(peak_properties, summary_path)
    return filename


# @profile
def rescale_intensity(
    i: float, rt: float, molecule: str, peak_properties: dict, isotopologue_lib: dict
):
    """Rescale intensity value for a given molecule according to scale factor and distribution function.

    Args:
        i (TYPE): Description
        rt (TYPE): Description
        molecule (TYPE): Description
        peak_properties (TYPE): Description
        isotopologue_lib (TYPE): Description

    Returns:
        TYPE: Description
    """
    scale_func = peak_properties[f"{molecule}"]["peak_function"]
    rt_max = (
        peak_properties[f"{molecule}"]["scan_start_time"]
        + peak_properties[f"{molecule}"]["peak_width"]
    )

    if scale_func == "gauss":
        mu = (
            peak_properties[f"{molecule}"]["scan_start_time"]
            + 0.5 * peak_properties[f"{molecule}"]["peak_width"]
        )
        dist_scale_factor = distributions[scale_func](
            rt,
            mu=mu,
            sigma=peak_properties[f"{molecule}"]["peak_params"].get(
                "sigma", peak_properties[f"{molecule}"]["peak_width"] / 10
            ),
        )
    elif scale_func == "gamma":
        dist_scale_factor = distributions[scale_func](
            rt,
            a=peak_properties[f"{molecule}"]["peak_params"]["a"],
            scale=peak_properties[f"{molecule}"]["peak_params"]["scale"],
        )
    elif scale_func == "gauss_tail":
        mu = (
            peak_properties[f"{molecule}"]["scan_start_time"]
            + 0.3 * peak_properties[f"{molecule}"]["peak_width"]
        )
        dist_scale_factor = distributions[scale_func](
            rt,
            mu=mu,
            sigma=0.12 * (rt - peak_properties[f"{molecule}"]["scan_start_time"]) + 2,
            scan_start_time=peak_properties[f"{molecule}"]["scan_start_time"],
        )
    elif scale_func is None:
        dist_scale_factor = 1
    # TODO use ionization_effiency here
    i *= (
        dist_scale_factor
        * peak_properties[f"{molecule}"].get("peak_scaling_factor", 1e3)
        * peak_properties[f"{molecule}"].get("ionization_effiency", 1)
    )
    return i


def generate_scans(
    isotopologue_lib: dict,
    peak_properties: dict,
    interval_tree: IntervalTree,
    fragmentor: AbstractFragmentor,
    noise_injector: AbstractNoiseInjector,
    mzml_params: dict,
):
    """Summary.

    Args:
        isotopologue_lib (TYPE): Description
        peak_properties (TYPE): Description
        fragmentation_function (A): Description
        mzml_params (TYPE): Description
    """
    logger.info("Initialize chimeric spectra counter")
    chimeric_count = 0
    chimeric = Counter()
    logger.info("Start generating scans")
    t0 = time.time()
    gradient_length = mzml_params["gradient_length"]
    ms_rt_diff = mzml_params.get("ms_rt_diff", 0.03)
    t: float = 0

    mol_scan_dict: Dict[str, Dict[str, list]] = {}
    scans: List[Tuple[Scan, List[Scan]]] = []
    # i: int = 0
    spec_id: int = 1
    de_tracker: Dict[str, int] = {}
    de_stats: dict = {}

    mol_scan_dict = {
        mol: {"ms1_scans": [], "ms2_scans": []} for mol in isotopologue_lib
    }
    molecules = list(isotopologue_lib.keys())

    progress_bar = tqdm(
        total=gradient_length,
        desc="Generating scans",
        bar_format="{desc}: {percentage:3.0f}%|{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}",
    )
    while t < gradient_length:
        scan_peaks: List[Tuple[float, float]] = []
        scan_peaks = {}
        mol_i = []
        mol_monoisotopic = {}
        candidates = interval_tree.at(t)
        # print(len(candidates))
        for mol in candidates:
            # if len(candidates) > 1:
            mol = mol.data
            mol_plus = f"{mol}"
            mz = np.array(isotopologue_lib[mol]["mz"])
            intensity = np.array(isotopologue_lib[mol]["i"])
            intensity = rescale_intensity(
                intensity, t, mol, peak_properties, isotopologue_lib
            )

            mask = intensity > mzml_params["min_intensity"]
            intensity = intensity[mask]

            # clip max intensity
            intensity = np.clip(
                intensity, a_min=None, a_max=mzml_params.get("max_intensity", 1e10)
            )

            mz = mz[mask]
            mol_peaks = list(zip(mz, intensity))
            mol_peaks = {round(mz, 6): _i for mz, _i in list(zip(mz, intensity))}
            # !FIXED! multiple molecules which share mz should have summed up intensityies for that shared mzs
            if len(mol_peaks) > 0:
                mol_i.append((mol, mz[0], sum(intensity)))
                # scan_peaks.extend(mol_peaks)
                for mz, intensity in mol_peaks.items():
                    if mz in scan_peaks:
                        scan_peaks[mz] += intensity
                    else:
                        scan_peaks[mz] = intensity
                mol_scan_dict[mol]["ms1_scans"].append(spec_id)
                highest_peak = max(mol_peaks.items(), key=lambda x: x[1])
                mol_monoisotopic[mol] = {
                    "mz": highest_peak[0],
                    "i": highest_peak[1],
                }
        scan_peaks = sorted(list(scan_peaks.items()), key=lambda x: x[1])
        if len(scan_peaks) > 0:
            mz, inten = zip(*scan_peaks)
        else:
            mz, inten = [], []

        s = Scan(
            {
                "mz": np.array(mz),
                "i": np.array(inten),
                "id": spec_id,
                "rt": t,
                "ms_level": 1,
            }
        )
        prec_scan_id = spec_id
        spec_id += 1

        sorting = s.mz.argsort()
        s.mz = s.mz[sorting]
        s.i = s.i[sorting]

        # add noise
        s = noise_injector.inject_noise(s)

        # i += 1
        scans.append((s, []))
        t += ms_rt_diff
        progress_bar.update(ms_rt_diff)

        if t > gradient_length:
            break

        fragment_spec_index = 0
        max_ms2_spectra = mzml_params.get("max_ms2_spectra", 10)
        if len(mol_i) < max_ms2_spectra:
            max_ms2_spectra = len(mol_i)
        ms2_scan = None
        mol_i = sorted(mol_i, key=lambda x: x[2], reverse=True)
        logger.debug(f"All molecules eluting: {len(mol_i)}")
        logger.debug(f"currently # fragment spectra {len(scans[-1][1])}")

        mol_i = [
            mol
            for mol in mol_i
            if (de_tracker.get(mol[0], None) is None)
            or (t - de_tracker[mol[0]]) > mzml_params["dynamic_exclusion"]
        ]
        logger.debug(f"All molecules eluting after DE filtering: {len(mol_i)}")
        while len(scans[-1][1]) != max_ms2_spectra:
            logger.debug(f"Frag spec index {fragment_spec_index}")
            if fragment_spec_index > len(mol_i) - 1:
                # we evaluated fragmentation for every potential mol
                # and all will be skipped
                logger.debug(f"All possible mol are skipped due to DE")
                break
            mol = mol_i[fragment_spec_index][0]
            _mz = mol_i[fragment_spec_index][1]
            _intensity = mol_i[fragment_spec_index][2]
            fragment_spec_index += 1
            mol_plus = f"{mol}"
            all_mols_in_mz_and_rt_window = [
                mol.data
                for mol in candidates
                if (
                    abs(isotopologue_lib[mol.data]["mz"][0] - _mz)
                    < mzml_params["isolation_window_width"]
                )
            ]
            if len(all_mols_in_mz_and_rt_window) > 1:
                chimeric_count += 1
                chimeric[len(all_mols_in_mz_and_rt_window)] += 1
            if mol is None:
                # dont add empty MS2 scans but have just a much scans as precursors
                breakpoint()
                ms2_scan = Scan(
                    {
                        "mz": np.array([]),
                        "i": np.array([]),
                        "rt": t,
                        "id": spec_id,
                        "precursor_mz": 0,
                        "precursor_i": 0,
                        "precursor_charge": 1,
                        "precursor_scan_id": prec_scan_id,
                        "ms_level": 2,
                    }
                )
                spec_id += 1
                t += ms_rt_diff
                progress_bar.update(ms_rt_diff)

                if t > gradient_length:
                    break
            elif (peak_properties[mol_plus]["scan_start_time"] <= t) and (
                (
                    peak_properties[mol_plus]["scan_start_time"]
                    + peak_properties[mol_plus]["peak_width"]
                )
                >= t
            ):
                # fragment all molecules in isolation and rt window
                # check if molecule needs to be fragmented according to dynamic_exclusion rule
                if (
                    de_tracker.get(mol, None) is None
                    or (t - de_tracker[mol]) > mzml_params["dynamic_exclusion"]
                ):
                    logger.debug("Generate Fragment spec")
                    de_tracker[mol] = t
                    if mol not in de_stats:
                        de_stats[mol] = {"frag_events": 0, "frag_spec_ids": []}
                    de_stats[mol]["frag_events"] += 1
                    de_stats[mol]["frag_spec_ids"].append(spec_id)
                    peaks = fragmentor.fragment(all_mols_in_mz_and_rt_window)
                    frag_mz = peaks[:, 0]
                    frag_i = peaks[:, 1]
                    ms2_scan = Scan(
                        {
                            "mz": frag_mz,
                            "i": frag_i,
                            "rt": t,
                            "id": spec_id,
                            "precursor_mz": mol_monoisotopic[mol]["mz"],
                            "precursor_i": mol_monoisotopic[mol]["i"],
                            "precursor_charge": peak_properties[mol]["charge"],
                            "precursor_scan_id": prec_scan_id,
                            "ms_level": 2,
                        }
                    )
                    spec_id += 1
                    ms2_scan.i = rescale_intensity(
                        ms2_scan.i, t, mol, peak_properties, isotopologue_lib
                    )
                    ms2_scan = noise_injector.inject_noise(ms2_scan)
                    ms2_scan.i *= 0.5
                else:
                    logger.debug(f"Skip {mol} due to dynamic exclusion")
                    continue
                t += ms_rt_diff
                progress_bar.update(ms_rt_diff)
                if t > gradient_length:
                    break
            else:
                logger.debug(f"Skip {mol} since not in RT window")
                continue
            if mol is not None:
                mol_scan_dict[mol]["ms2_scans"].append(spec_id)
            if ms2_scan is None:
                # there are molecules in mol_i
                # however all molecules are excluded from fragmentation_function
                # => Don't do a scan and break the while loop
                # => We should rather continue and try to fragment the next mol!
                logger.debug(f"Continue and fragment next mol since MS2 scan is None")
                continue
            if (
                len(ms2_scan.mz) > -1
            ):  # TODO -1 to also add empty ms2 specs; 0 breaks tests currently ....
                sorting = ms2_scan.mz.argsort()
                ms2_scan.mz = ms2_scan.mz[sorting]
                ms2_scan.i = ms2_scan.i[sorting]
                logger.debug(f"Append MS2 scan with {mol}")
                scans[-1][1].append(ms2_scan)
    progress_bar.close()
    t1 = time.time()
    logger.info("Finished generating scans")
    logger.info(f"Generating scans took {t1-t0:.2f} seconds")
    logger.info(f"Found {chimeric_count} chimeric scans")

    return scans, mol_scan_dict


# @profile
def generate_molecule_isotopologue_lib(
    peak_properties: Dict[str, dict],
    charges: List[int] = None,
    trivial_names: Dict[str, str] = None,
):
    """Summary.

    Args:
        molecules (TYPE): Description
    """
    logger.info("Generate Isotopolgue Library")
    start = time.time()
    duplicate_formulas: Dict[str, List[str]] = {}
    for key in peak_properties:
        duplicate_formulas.setdefault(
            peak_properties[key]["chemical_formula"], []
        ).append(key)
    if charges is None:
        charges = [1]
    if len(peak_properties) > 0:
        molecules = [d["chemical_formula"] for d in peak_properties.values()]
        lib = pyqms.IsotopologueLibrary(
            molecules=molecules,
            charges=charges,
            verbose=False,
            trivial_names=trivial_names,
        )
        reduced_lib = {}
        # TODO fix to  support multiple charge states
        for mol in molecules:
            formula = lib.lookup["molecule to formula"][mol]
            data = lib[formula]["env"][(("N", "0.000"),)]
            for triv in lib.lookup["formula to trivial name"][formula]:
                reduced_lib[triv] = {
                    "mz": data[peak_properties[triv]["charge"]]["mz"],
                    "i": data["relabun"],
                }
    else:
        reduced_lib = {}
    tmp = {}
    for mol in reduced_lib:
        cc = peak_properties[mol]["chemical_formula"]
        for triv in duplicate_formulas[cc]:
            if triv not in reduced_lib:
                tmp[triv] = reduced_lib[mol]
    reduced_lib.update(tmp)
    logger.info(
        f"Generating IsotopologueLibrary took {(time.time() - start)/60} minutes"
    )
    return reduced_lib


# @profile
def write_scans(
    file: Union[str, io.TextIOWrapper], scans: List[Tuple[Scan, List[Scan]]]
) -> None:
    """Generate given scans to mzML file.

    Args:
        file (Union[str, io.TextIOWrapper]): Description
        scans (List[Tuple[Scan, List[Scan]]]): Description

    Returns:
        None: Description
    """
    t0 = time.time()
    logger.info("Start writing Scans")
    ms1_scans = len(scans)
    ms2_scans = 0
    ms2_scan_list = []
    for s in scans:
        ms2_scans += len(s[1])
        ms2_scan_list.append(len(s[1]))
    logger.info("Write {0} MS1 and {1} MS2 scans".format(ms1_scans, ms2_scans))
    id_format_str = "controllerType=0 controllerNumber=1 scan={i}"
    with MzMLWriter(file) as writer:
        # Add default controlled vocabularies
        writer.controlled_vocabularies()
        writer.format()
        # Open the run and spectrum list sections
        time_array = []
        intensity_array = []
        with writer.run(id="Simulated Run"):
            spectrum_count = len(scans) + sum([len(products) for _, products in scans])
            with writer.spectrum_list(count=spectrum_count):
                for scan, products in scans:
                    # Write Precursor scan
                    try:
                        index_of_max_i = np.argmax(scan.i)
                        max_i = scan.i[index_of_max_i]
                        mz_at_max_i = scan.mz[index_of_max_i]
                    except ValueError:
                        mz_at_max_i = 0
                        max_i = 0
                    spec_tic = sum(scan.i)
                    writer.write_spectrum(
                        scan.mz,
                        scan.i,
                        id=id_format_str.format(i=scan.id),
                        params=[
                            "MS1 Spectrum",
                            {"ms level": 1},
                            {
                                "scan start time": scan.retention_time,
                                "unit_name": "seconds",
                            },
                            {"total ion current": spec_tic},
                            {"base peak m/z": mz_at_max_i, "unitName": "m/z"},
                            {
                                "base peak intensity": max_i,
                                "unitName": "number of detector counts",
                            },
                        ],
                    )
                    time_array.append(scan.retention_time)
                    intensity_array.append(spec_tic)
                    # Write MSn scans
                    for prod in products:
                        writer.write_spectrum(
                            prod.mz,
                            prod.i,
                            id=id_format_str.format(i=prod.id),
                            params=[
                                "MSn Spectrum",
                                {"ms level": 2},
                                {
                                    "scan start time": prod.retention_time,
                                    "unit_name": "seconds",
                                },
                                {"total ion current": sum(prod.i)},
                            ],
                            precursor_information={
                                "mz": prod.precursor_mz,
                                "intensity": prod.precursor_i,
                                "charge": prod.precursor_charge,
                                "scan_id": id_format_str.format(i=scan.id),
                                "spectrum_reference": id_format_str.format(i=scan.id),
                                "activation": ["HCD", {"collision energy": 25.0}],
                            },
                        )
            with writer.chromatogram_list(count=1):
                writer.write_chromatogram(
                    time_array,
                    intensity_array,
                    id="TIC",
                    chromatogram_type="total ion current",
                )
    t1 = time.time()
    logger.info(f"Writing mzML took {(t1-t0)/60:.2f} minutes")
    return



"""Mapping nucleoside names to its fragments."""

KB_FRAGMENTATION_INFO: dict = {
    "uridine": {
        # U
        "comments": "precursor can not be seen in MS2 in some cases",
        "references": ["Jora et al. 2018"],
        "fragments": {
            "uridine -Ribose": {"formula": "C(4)H(4)N(2)O(2)"},  # 113.0345538436
            "uridine -Ribose -N(1)H(3)": {
                "formula": "C(4)H(1)N(1)O(2)",  # 96.0080047430
                "hcd": True,
            },
        },
        "precursors": {
            "uridine": {},
            "uridine dimer": {},
        },
    },
    "5-methyluridine": {
        # m5U
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "5-methyluridine": {},
            "5-methyluridine +N(1)H(2)": {"formula": "C(10)H(16)N(3)O(6)"},
            "5-methyluridine dimer": {},
        },
        "fragments": {
            "5-methyluridine -Ribose": {"formula": "C(5)H(6)N(2)O(2)"},
            "5-methyluridine -Ribose -N(1)H(3)": {
                "formula": "C(5)H(3)N(1)O(2)",  # 110.0236548074
                "hcd": True,
            },
            "5-methyluridine -Ribose -H(2)O(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
                "hcd": True,
            },
            # '5-methyluridine -Ribose -C(1)H(1)N(1)O(1)' : {
            #     'formula': 'C(4)H(5)N(1)O(1)', #84.04
            #     'hcd':True
            # },
            # '5-methyluridine -Ribose -C(1)H(3)N(1)O(1)' : {
            #     'formula': 'C(4)H(3)N(1)O(1)', #82.028
            #     'hcd':True
            # },
            # '5-methyluridine -Ribose -C(2)H(2)N(2)O(1)' : {
            #     'formula': 'C(3)H(4)O(1)', #57.033
            #     'hcd':True
            # },
            # '5-methyluridine -Ribose -C(2)H(1)N(1)O(2)' : {
            #     'formula': 'C(3)H(5)N(1)', #56.050
            #     'hcd':True
            # },
            # '5-methyluridine -Ribose -C(2)H(3)N(1)O(2)' : {
            #     'formula': 'C(3)H(3)N(1)', #53.034
            #     'hcd':True
            # },
        },
        "exclusion_fragments": {
            "5-methyluridine -Methylated ribose": {
                "formula": "C(4)H(4)N(2)O(2)"  # 113.0345538436
            }
        },
    },
    "3-methyluridine": {
        # m3U
        "comments": "Exclusion fragments are partly m5U specific fragments",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "3-methyluridine": {},
            "3-methyluridine +N(1)H(2)": {"formula": "C(10)H(16)N(3)O(6)"},
            "3-methyluridine dimer": {},
        },
        "fragments": {
            "3-methyluridine -Ribose": {"formula": "C(5)H(6)N(2)O(2)"},
            "3-methyluridine -Ribose -C(1)H(5)N(1)": {
                "formula": "C(4)H(1)N(1)O(2)",  # 96
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "3-methyluridine -Methylated ribose": {
                "formula": "C(4)H(4)N(2)O(2)"  # 113.0345538436
            },
            "3-methyluridine -Ribose -N(1)H(3)": {
                "formula": "C(5)H(3)N(1)O(2)",  # 110.02
                "hcd": True,
            },
            "3-methyluridine -Ribose -H(2)O(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.03
                "hcd": True,
            },
        },
    },
    "2′-O-methyluridine": {
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {"2′-O-methyluridine": {}},
        "fragments": {
            "2′-O-methyluridine -Methylated ribose": {
                "formula": "C(4)H(4)N(2)O(2)"  # 113.035
            },
            "2′-O-methyluridine -Methylated ribose -H(3)N(1)": {
                "formula": "C(4)H(1)N(1)O(2)",  # 96.008
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "2′-O-methyluridine -Ribose": {"formula": "C(5)H(6)N(2)O(2)"}  # 127
        },
    },
    "2′-O-methylpseudouridine": {
        "comments": "Verification required",
        "references": [],
        "precursors": {"2′-O-methylpseudouridine": {}},
        "fragments": {
            "2′-O-methylpseudouridine -H(4)O(4)": {"formula": "C(10)H(10)N(2)O(4)"}
        },
    },
    "pseudouridine": {
        "references": ["Dudley et al. 2005", "Pomerantz et al. 2005"],
        "comments": "peak 209.0556832124: ribose is doubly dehydrated; peak 191.0451185280: ribose is triply dehydrated",
        "fragments": {
            "pseudouridine -H(4)O(2)": {
                "formula": "C(9)H(8)N(2)O(4)",  # 209.0556832124
                "hcd": False,
            },
            "pseudouridine -H(6)O(3)": {
                "formula": "C(9)H(6)N(2)O(3)",  # 191.0451185280
                "hcd": False,
            },
            "pseudouridine -C(3)H(6)O(3)": {
                "formula": "C(6)H(6)N(2)O(3)",  # 155.0451185280
                "hcd": False,
            },
            "pseudouridine -C(4)H(8)O(4)": {
                "formula": "C(5)H(4)N(2)O(2)",  # 125.0345538436
                "hcd": True,
            },
            "pseudouridine -C(5)H(7)N(1)O(4)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
            "pseudouridine -C(5)H(9)N(1)O(5)": {
                "formula": "C(4)H(3)N(1)O(1)",  # 82.0287401874
                "hcd": True,
            },
            "pseudouridine -C(5)H(6)N(1)O(6)": {
                "formula": "C(4)H(5)N(1)",  # 68.0494756318
                "hcd": True,
            },
        },
        "precursors": {
            "pseudouridine K adduct": {"formula": "C(9)H(11)N(2)O(6)K(1)"},
            "pseudouridine": {},
            "pseudouridine +N(1)H(2)": {"formula": "C(9)H(14)N(3)O(6"},
            # 'pseudouridine dimer' : {},
            # 'pseudouridine K adduct' : {
            # 'formula' : 'C(9)H(11)N(2)O(6)K(1)'
            # },
            # 'pseudouridine +Ammonia' : {
            # 'formula' : 'C(9)H(15)N(3)O(6)'
            # },
        },
    },
    "1-methylpseudouridine": {
        # m1Y
        "references": ["Dudley et al. 2005"],
        "comments": "Peaks similar to pseudouridine with addtional +C(1)H(2)",
        "fragments": {
            "1-methylpseudouridine -H(4)O(2)": {
                "formula": "C(10)H(10)N(2)O(4)",  # 223.0713332768
                "hcd": False,
            },
            "1-methylpseudouridine -H(6)O(3)": {
                "formula": "C(10)H(8)N(2)O(3)",  # 205.0607685924
                "hcd": False,
            },
            "1-methylpseudouridine -C(3)H(6)O(3)": {
                "formula": "C(7)H(8)N(2)O(3)",  # 169.0607685924
                "hcd": False,
            },
            "1-methylpseudouridine -C(4)H(8)O(4)": {
                "formula": "C(6)H(6)N(2)O(2)",  # 139.0502039080
                "hcd": True,
            },
            "1-methylpseudouridine -C(5)H(9)N(1)O(5)": {
                "formula": "C(5)H(5)N(1)O(1)",  # 96.0443902518
                "hcd": True,
            },
        },
        "precursors": {
            "1-methylpseudouridine K adduct": {},
            "1-methylpseudouridine": {},
            "1-methylpseudouridine dimer": {},
        },
    },
    "3-methylpseudouridine": {
        # m3Y
        "references": ["Dudley et al. 2005"],
        "comments": "Verification required; Peaks similar to pseudouridine with addtional +C(1)H(2)",
        "fragments": {
            "3-methylpseudouridine -H(4)O(2)": {
                "formula": "C(10)H(10)N(2)O(4)",  # 223.0713332768
                "hcd": False,
            },
            "3-methylpseudouridine -H(6)O(3)": {
                "formula": "C(10)H(8)N(2)O(3)",  # 205.0607685924
                "hcd": False,
            },
            "3-methylpseudouridine -C(3)H(6)O(3)": {
                "formula": "C(7)H(8)N(2)O(3)",  # 169.0607685924
                "hcd": False,
            },
            "3-methylpseudouridine -C(4)H(8)O(4)": {
                "formula": "C(6)H(6)N(2)O(2)",  # 139.0502039080
                "hcd": True,
            },
            "3-methylpseudouridine -C(5)H(9)N(1)O(5)": {
                "formula": "C(5)H(5)N(1)O(1)",  # 96.0443902518
                "hcd": True,
            },
        },
        "precursors": {
            "3-methylpseudouridine K adduct": {},
            "3-methylpseudouridine": {},
            "3-methylpseudouridine dimer": {},
        },
    },
    "inosine": {
        # I
        "references": [],
        "comments": "Precursor ion can not be seen in MS2 in most cases",
        "fragments": {
            "inosine -Ribose": {"formula": "C(5)H(4)N(4)O(1)"}  # 137.0457872316
        },
        "precursors": {"inosine": {}, "inosine dimer": {}},
    },
    "1-methylinosine": {
        # m1I
        "references": [],
        "comments": "",
        "precursors": {"1-methylinosine": {}, "1-methylinosine dimer": {}},
        "fragments": {
            "1-methylinosine -Ribose": {
                "formula": "C(6)H(6)N(4)O(1)"  # 151.06143729597002
            }
        },
        "exclusion_fragments": {
            "1-methylinosine -Methylated ribose": {
                "formula": "C(5)H(4)N(4)O(1)"  # 137.04578723157002
            }
        },
    },
    "2′-O-methylinosine": {
        # Im
        "references": [],
        "comments": "",
        "precursors": {"2′-O-methylinosine": {}, "2′-O-methylinosine dimer": {}},
        "fragments": {
            "2′-O-methylinosine -Methylated ribose": {
                "formula": "C(5)H(4)N(4)O(1)"  # 137.04578723157002
            }
        },
        "exclusion_fragments": {
            "2′-O-methylinosine -Ribose": {
                "formula": "C(6)H(6)N(4)O(1)"  # 151.06143729597002
            }
        },
    },
    "N6-isopentenyladenosine": {
        # i6A
        "references": ["modomics.org"],
        "comments": "",
        "precursors": {"N6-isopentenyladenosine": {}},
        "fragments": {
            "N6-isopentenyladenosine -Ribose": {
                "formula": "C(10)H(13)N(5)",  # 204.1243719054
                "hcd": False,
            },
            "N6-isopentenyladenosine -Ribose -C(4)H(8)": {
                "formula": "C(6)H(5)N(5)",  # 148.0617716478
                "hcd": True,
            },
            "N6-isopentenyladenosine -Ribose -C(5)H(8)": {
                "formula": "C(5)H(5)N(5)",  # 136.0617716478
                "hcd": True,
            },
        },
    },
    "N6-formyladenosine": {
        # f6A
        "references": ["Fu et al. 2013"],
        "comments": "Verification by standard required",
        "precursors": {"N6-formyladenosine": {}},
        "fragments": {
            "N6-formyladenosine -Ribose": {
                "formula": "C(6)H(5)N(5)O(1)",  # 164.05668626777
                "hcd": False,
            },
            "N6-formyladenosine -Ribose -C(1)O(1)": {
                "formula": "C(5)H(5)N(5)",  # 136.0617716478
                "hcd": True,
            },
        },
    },
    "5-methoxyuridine": {
        # mo5U
        "comment": "Can be a co-eluting fragment (in-source?) of cm5U but with different fragments",
        "references": ["Nelson and McCloskey 1994"],
        "fragments": {
            "5-methoxyuridine -Ribose": {
                "formula": "C(5)H(6)N(2)O(3)",  # 143.0451185280
                "hcd": False,
            },
            # "5-methoxyuridine -Ribose -N(1)H(3)": {
            #     "formula": "C(5)H(3)N(1)O(3)",  # 126.0185694274
            #     "hcd": True,
            # },
        },
        "precursors": {
            "5-methoxyuridine": {},
            "5-methoxyuridine dimer": {},
            "5-methoxyuridine K adduct": {},
        },
    },
    "N6-threonylcarbamoyladenosine": {
        # t6A
        "comments": "120, 136, 162 and 281 are major peaks",
        "references": ["Nagao et al. 2017"],
        "fragments": {
            # 'N6-threonylcarbamoyladenosine -Ribose' : {
            #     'formula' : 'C(10)H(12)N(6)O(4)' # 281.0992793572
            # },
            "N6-threonylcarbamoyladenosine -Ribose -C(4)H(9)N(1)O(3)": {
                "formula": "C(6)H(3)N(5)O(1)"  # 162.0410362034
            },
            "N6-threonylcarbamoyladenosine -Ribose -C(5)H(7)N(1)O(4)": {
                "formula": "C(5)H(5)N(5)"  # 136.0617716478
            },
            "N6-threonylcarbamoyladenosine -Ribose -C(6)H(3)N(5)O(1)": {
                "formula": "C(4)H(9)N(1)O(3)",  # 120.0655196206
                "hcd": False,
            },
        },
        "precursors": {
            "N6-threonylcarbamoyladenosine": {
                "formula": "C(15)H(20)N(6)O(8)"  # 413.1415380948
            }
        },
    },
    "cyclic N6-threonylcarbamoyladenosine": {
        "references": ["Myauchi et al. 2013"],
        "comments": "",
        "precursors": {"cyclic N6-threonylcarbamoyladenosine": {}},
        "fragments": {
            "cyclic N6-threonylcarbamoyladenosine -Ribose": {
                "formula": "C(10)H(10)N(6)O(3)"  # 263.0887146728
            },
            "cyclic N6-threonylcarbamoyladenosine -Ribose -C(2)H(4)O(1)": {
                "formula": "C(8)H(6)N(6)O(2)"  # 219.0624999240
            },
            "cyclic N6-threonylcarbamoyladenosine -Ribose -C(4)H(7)N(1)O(2)": {
                "formula": "C(6)H(3)N(5)O(1)"  # 162.0410362034
            },
        },
    },
    "guanosine": {
        "references": [],
        "comments": "",
        "precursors": {
            "guanosine dimer": {"formula": "C(20)H(26)N(10)O(10)"},
            "guanosine": {"formula": "C(10)H(13)N(5)O(5)"},
            "guanosine K adduct": {"formula": "C(10)H(12)K(1)N(5)O(5)"},
        },
        "fragments": {
            # 'guanosine': {
            #   'formula': 'C(10)H(13)N(5)O(5)'
            # },
            "guanosine -Ribose": {"formula": "C(5)H(5)N(5)O(1)"},  # 152.0566862678
            "guanosine -Ribose -H(1)N(1) +O(1)": {
                "formula": "C(5)H(4)N(4)O(2)",  # 153.0407018516
                "hcd": True,
            },
            "guanosine -Ribose -H(3)N(1)": {
                "formula": "C(5)H(2)N(4)O(1)",  # 135.0301371672
                "hcd": True,
            },
        },
    },
    "cytidine": {
        "references": [],
        "comments": "",
        "precursors": {
            "cytidine dimer": {},
            "cytidine": {},
        },
        "fragments": {
            # 'cytidine': {
            #     'formula': 'C(9)H(13)N(3)O(5)'
            # },
            "cytidine -Ribose": {"formula": "C(4)H(5)N(3)O(1)"}  # 112.0505382598
        },
    },
    "3-methylcytidine": {
        # m3C
        "comments": "No dimer is formed! Methyl group blocks dimer formation?",
        "references": ["Jora et al. 2018"],
        "precursors": {
            #
            # '3-methylcytidine dimer' : { 'formula' : '' },
            # 'cytidine dimer semi-labeled' : { 'formula' : '' },
            "3-methylcytidine": {"formula": ""}
        },
        "fragments": {
            "3-methylcytidine -Ribose": {
                "formula": "C(5)H(7)N(3)O(1)"  # 126.0661883242
            },
            "3-methylcytidine -Ribose -H(3)N(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
                "hcd": True,
            },
            "3-methylcytidine -Ribose -C(1)H(5)N(1)": {
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "3-methylcytidine -Methylated ribose": {
                "formula": "C(4)H(5)N(3)O(1)",  # 112.0505382598
                # 'hcd' : True
            },
            # '3-methylcytidine -Ribose' : {
            #     'formula': 'C(5)H(7)N(3)O(1)',
            #     'hcd': True
            # },
        },
        "dimer_possible": False,
    },
    "5-methylcytidine": {
        # m5C
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "5-methylcytidine dimer": {},
            "5-methylcytidine": {},
        },
        "fragments": {
            "5-methylcytidine -Ribose": {
                "formula": "C(5)H(7)N(3)O(1)"  # 126.0661883242
            },
            "5-methylcytidine -Ribose -H(3)N(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
                "hcd": True,
            },
            "5-methylcytidine -Ribose -H(2)O(1)": {
                "formula": "C(5)H(5)N(3)",  # 108.0556236398
                "hcd": True,
            },
            # '5-methylcytidine -Ribose -C(1)H(1)N(1)O(1)' : {
            #     'formula': 'C(4)H(6)N(2)', # 83.0603746680
            #     'hcd': True
            # },
        },
        "exclusion_fragments": {
            "5-methylcytidine -Methylated ribose": {
                "formula": "C(4)H(5)N(3)O(1)",  # 112.0505382598
                "hcd": False,
            },
            "5-methylcytidine -Ribose -C(1)H(5)N(1)": {
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            },
        },
    },
    "N4-methylcytidine": {
        # m4C
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "N4-methylcytidine dimer": {},
            "N4-methylcytidine": {},
        },
        "fragments": {
            "N4-methylcytidine -Ribose": {
                "formula": "C(5)H(7)N(3)O(1)"  # 126.0661883242
            },
            "N4-methylcytidine -Ribose -H(3)N(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
                "hcd": True,
            },
            "N4-methylcytidine -Ribose -H(2)O(1)": {
                "formula": "C(5)H(5)N(3)",  # 108.0556236398
                "hcd": True,
            },
            "N4-methylcytidine -Ribose -C(1)H(5)N(1)": {
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            },
            "N4-methylcytidine -Ribose -C(1)H(1)N(1)O(1)": {
                "formula": "C(4)H(6)N(2)",  # 83.0603746680
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "N4-methylcytidine -Methylated ribose": {
                "formula": "C(4)H(5)N(3)O(1)"  # 112.0505382598
            }
        },
    },
    "2′-O-methylcytidine": {
        # Cm
        "comments": "",
        "references": [],
        "precursors": {"2′-O-methylcytidine": {"formula": ""}},
        "fragments": {
            "2′-O-methylcytidine -Methylated ribose": {
                "formula": "C(4)H(5)N(3)O(1)"  # 112.0505382598
            },
            "2′-O-methylcytidine -Methylated ribose -H(3)N(1)": {
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            },
            # '2′-O-methylcytidine -Methylated ribose -C(1)H(1)N(1)O(1)' : {
            #     'formula': 'C(3)H(4)N(2)', # 69.044
            #     'hcd': True
            # },
        },
        "exclusion_fragments": {
            "2′-O-methylcytidine -Ribose": {
                "formula": "C(5)H(7)N(3)O(1)"  # 126.0661883242
            }
        },
    },
    "5-methoxycarbonylmethyluridine": {
        # mcm5U
        "comments": "",
        "references": [],
        "fragments": {
            "5-methoxycarbonylmethyluridine -Ribose": {
                "formula": "C(7)H(8)N(2)O(4)",  # 185.0556832124
                "hcd": False,
            },
            "5-methoxycarbonylmethyluridine -Ribose -C(1)H(4)O(1)": {
                "formula": "C(6)H(4)N(2)O(3)",  # 153.0294684636
                "hcd": False,
            },
            "5-methoxycarbonylmethyluridine -Ribose -C(2)H(4)O(2)": {
                "formula": "C(5)H(4)N(2)O(2)"  # 125.0345538436
            },
            "5-methoxycarbonylmethyluridine -Ribose -C(3)H(3)N(1)O(2)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
            # '5-methoxycarbonylmethyluridine -Ribose -C(3)H(5)N(1)O(1)' : {
            #     'formula': 'C(4)H(3)N(1)O(1)', # 82.0287401874
            #     'hcd':True
            # },
            # '5-methoxycarbonylmethyluridine -Ribose -C(4)H(5)N(1)O(2)' : {
            #     'formula': 'C(3)H(3)N(1)', # 54.0338255674
            #     'hcd':True
            # },
        },
        "precursors": {
            "5-methoxycarbonylmethyluridine": {},
            "5-methoxycarbonylmethyluridine +H(3)N(1)": {
                "formula": "C(12)H(19)N(3)O(8)"
            },
            "5-methoxycarbonylmethyluridine dimer": {},
            "5-methoxycarbonylmethyluridine dimer +H(3)N(1)": {
                "formula": "C(24)H(35)N(5)O(16)"
            },
        },
    },
    "5-methoxycarbonylmethyl-2-thiouridine": {
        # mcm5s2U
        "comments": "",
        "references": [],
        "fragments": {
            "5-methoxycarbonylmethyl-2-thiouridine -Ribose": {
                "formula": "C(7)H(8)N(2)O(3)S(1)",  # 201.0328397664
                "hcd": False,
            },
            "5-methoxycarbonylmethyl-2-thiouridine -Ribose -C(1)H(4)O(1)": {
                "formula": "C(6)H(4)N(2)O(2)S(1)",  # 169.0066250176
                "hcd": False,
            },
            "5-methoxycarbonylmethyl-2-thiouridine -Ribose -C(2)H(4)O(2)": {
                "formula": "C(5)H(4)N(2)O(1)S(1)",  # 141.0117103976
                "hcd": True,
            },
            "5-methoxycarbonylmethyl-2-thiouridine -Ribose -C(3)H(3)N(1)O(1)S(1)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
        },
        "precursors": {"5-methoxycarbonylmethyl-2-thiouridine": {}},
    },
    "5-carbamoylmethyluridine": {
        # ncm5U
        "comments": "",
        "references": [],
        "precursors": {"5-carbamoylmethyluridine": {}},
        "fragments": {
            "5-carbamoylmethyluridine -Ribose": {
                "formula": "C(6)H(7)N(3)O(3)",  # 170.0560175642
                "hcd": False,
            },
            "5-carbamoylmethyluridine -Ribose -N(1)H(3)": {
                "formula": "C(6)H(4)N(2)O(3)",  # 153.0294684636
                "hcd": False,
            },
            "5-carbamoylmethyluridine -Ribose -C(1)H(3)N(1)O(1)": {
                "formula": "C(5)H(4)N(2)O(2)"  # 125.0345538436
            },
            "5-carbamoylmethyluridine -Ribose -C(2)H(2)N(2)O(1)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
            "5-carbamoylmethyluridine -Ribose -C(2)H(4)N(2)O(2)": {
                "formula": "C(4)H(3)N(1)O(1)",  # 82.0287401874
                "hcd": True,
            },
        },
    },
    "5-methoxycarbonylmethyl-2′-O-methyluridine": {
        # mcm5Um
        "comments": "",
        "references": [],
        "precursors": {
            "5-methoxycarbonylmethyl-2′-O-methyluridine": {
                "formula": "C(13)H(18)N(2)O(8)"  # 331.1135920144
            }
        },
        "fragments": {
            "5-methoxycarbonylmethyl-2′-O-methyluridine -Methylated ribose": {
                "formula": "C(7)H(8)N(2)O(4)",  # 185.0556832124
                "hcd": False,
            },
            "5-methoxycarbonylmethyl-2′-O-methyluridine -Methylated ribose -C(2)H(4)O(2)": {
                "formula": "C(5)H(4)N(2)O(2)"  # 125.0345538436
            },
            "5-methoxycarbonylmethyl-2′-O-methyluridine -Methylated ribose -C(3)H(3)N(1)O(2)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
            "5-methoxycarbonylmethyl-2′-O-methyluridine -Methylated ribose -C(3)H(4)N(1)O(3)": {
                "formula": "C(4)H(3)N(1)O(1)",  # 82.0287401874
                "hcd": True,
            },
        },
    },
    "5-carboxymethyluridine": {
        # cm5U
        "comments": "",
        "references": [],
        "precursors": {"5-carboxymethyluridine": {}},
        "fragments": {
            "5-carboxymethyluridine -Ribose": {
                "formula": "C(6)H(6)N(2)O(4)",  # 171.0400331480
                "hcd": False,
            },
            "5-carboxymethyluridine -Ribose -H(2)O(1)": {
                "formula": "C(6)H(4)N(2)O(3)",  # 153.0294684636
                "hcd": False,
            },  # 153
            "5-carboxymethyluridine -Ribose -C(1)H(2)O(2)": {
                "formula": "C(5)H(4)N(2)O(2)"  # 125.0345538436
            },
            "5-carbamoylmethyluridine -Ribose -C(2)H(2)N(2)O(1)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
            "5-carbamoylmethyluridine -Ribose -C(2)H(4)N(2)O(2)": {
                "formula": "C(4)H(3)N(1)O(1)",  # 82.0287401874
                "hcd": True,
            },
        },
    },
    "5-carbamoylmethyl-2-thiouridine": {
        # ncm5s2U
        "comments": "",
        "references": [],
        "precursors": {"5-carbamoylmethyl-2-thiouridine": {}},
        "fragments": {
            "5-carbamoylmethyl-2-thiouridine -Ribose": {
                "formula": "C(6)H(7)N(3)O(2)S(1)",  # 186.0331741182
                "hcd": False,
            },
            "5-carbamoylmethyl-2-thiouridine -Ribose -H(3)N(1)": {
                "formula": "C(6)H(4)N(2)O(2)S(1)",  # 169.0066250176
                "hcd": False,
            },
            "5-carbamoylmethyl-2-thiouridine -Ribose -C(1)H(3)O(1)N(1)": {
                "formula": "C(5)H(4)N(2)O(1)S(1)"  # 141.0117103976
            },
            "5-carbamoylmethyl-2-thiouridine -Ribose -C(2)H(2)N(2)O(1)S(1)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
            "5-carbamoylmethyl-2-thiouridine -Ribose -C(2)H(4)N(2)O(2)S(1)": {
                "formula": "C(4)H(3)N(1)O(1)",  # 82.0287401874
                "hcd": True,
            },
        },
    },
    "5-carboxymethyl-2-thiouridine": {
        # cm5s2U
        "comments": "",
        "references": [],
        "precursors": {"5-carboxymethyl-2-thiouridine": {}},
        "fragments": {
            "5-carboxymethyl-2-thiouridine -Ribose": {
                "formula": "C(6)H(6)N(2)O(3)S(1)",  # 187.01718970197
                "hcd": False,
            },
            "5-carboxymethyl-2-thiouridine -Ribose -C(1)H(2)O(2)": {
                "formula": "C(5)H(4)N(2)O(1)S(1)"  # 141.0117103976
            },  # 141
            "5-carboxymethyl-2-thiouridine -Ribose -C(2)H(1)N(2)O(1)S(1)": {
                "formula": "C(4)H(5)N(1)O(2)",  # 100.0393048718
                "hcd": True,
            },
            "5-carboxymethyl-2-thiouridine -Ribose -C(2)H(3)N(1)O(2)S(1)": {
                "formula": "C(4)H(3)N(1)O(1)",  # 82.0287401874
                "hcd": True,
            },
        },
    },
    "5-carbamoylhydroxymethyluridine": {
        # nchm5U
        "comments": "",
        "references": [],
        "precursors": {"5-carbamoylhydroxymethyluridine": {}},
        "fragments": {
            "5-carbamoylhydroxymethyluridine -Ribose": {
                "formula": "C(6)H(7)N(3)O(4)",  # 186.0509321842
                "hcd": False,
            },
            "5-carbamoylhydroxymethyluridine -Ribose -C(1)H(3)N(1)O(1)": {
                "formula": "C(5)H(4)N(2)O(3)",  # 141.0294684636
                "hcd": True,
            },
            "5-carbamoylhydroxymethyluridine -Ribose -C(1)H(2)O(2)": {
                "formula": "C(5)H(5)N(3)O(2)",  # 140.0454528798
                "hcd": True,
            },
            "5-carbamoylhydroxymethyluridine -Ribose -C(1)H(2)O(2)": {
                "formula": "C(4)H(4)N(2)O(1)",  # 97.039639
                "hcd": True,
            },
            "5-carbamoylhydroxymethyluridine -Ribose -C(1)H(2)O(2)": {
                "formula": "C(3)H(3)N(1)O(1)",  # 70.0287401874
                "hcd": True,
            },
        },
    },
    "5-methyl-2-thiouridine": {
        # m5s2U
        "comments": "",
        "references": [],
        "precursors": {"5-methyl-2-thiouridine": {"formula": "C(10)H(14)N(2)O(5)S(1)"}},
        "fragments": {
            "5-methyl-2-thiouridine -Ribose": {
                "formula": "C(5)H(6)N(2)O(1)S(1)"  # 143.0273604620
            },
            "5-methyl-2-thiouridine -Ribose -N(1)H(3)": {
                "formula": "C(5)H(3)N(1)O(1)S(1)",  # 126.0008113614
                "hcd": True,
            },
        },
    },
    "5-aminomethyl-2-thiouridine": {
        # nm5s2U
        "comments": "",
        "references": [],
        "precursors": {
            "5-aminomethyl-2-thiouridine": {"formula": "C(10)H(15)N(3)O(5)S(1)"}
        },
        "fragments": {
            "5-aminomethyl-2-thiouridine -Ribose": {
                "formula": "C(5)H(7)N(3)O(1)S(1)"  # 158.038663833319
            },
            # "5-aminomethyl-2-thiouridine -Ribose -H(1)S(1)": {
            #     "formula": "C(5)H(6)N(3)O(1)",# 125.05836329197
            #     "hcd" : True
            #     not verified!!!
            # },
        },
    },
    "5-hydroxyuridine": {
        # ho5U
        "comments": "",
        "references": ["Nelson and McCloskey 1994"],
        "precursors": {"5-hydroxyuridine": {}},
        "fragments": {
            "5-hydroxyuridine -Ribose": {
                "formula": "C(4)H(4)N(2)O(3)"  # 129.0294684636
            },
            # "5-hydroxyuridine -Ribose -N(1)H(3)": {
            #     "formula": "C(4)H(1)N(1)O(3)", # 112.0029193630
            #     "hcd":True
            # },
            # "5-hydroxyuridine -Ribose -C(1)H(3)N(1)O(1)": {
            #     "formula": "C(3)H(1)N(1)O(2)", # 84.0080047430
            #     "hcd":True
            # },
        },
    },
    # "5,2′-O-dimethyluridine": {
    #     # m5Um
    #     "comments" : "Fragments similar to m5U (less similar to m3U, Verification required",
    #     "references" : ["Jora et al. 2018"],
    #     "precursors": {
    #         "5,2′-O-dimethyluridine": {
    #             "formula": "C(11)H(16)N(2)O(6)" # 273.1081127100
    #         }
    #     },
    #     "fragments": {
    #         "5,2′-O-dimethyluridine -Ribose": {
    #             "formula": "C(6)H(8)N(2)O(2)"  # 141.06585397237
    #         },
    #         "5,2′-O-dimethyluridine -Methylated ribose -N(1)H(3)": {
    #             "formula": "C(5)H(3)N(1)O(2)",  # 110.0236548074
    #             "hcd": True,
    #         },
    #         "5,2′-O-dimethyluridine -Methylated ribose -H(2)O(1)": {
    #             "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
    #             "hcd": True,
    #         },
    #         # '5,2′-O-dimethyluridine -Ribose -C(1)H(1)N(1)O(1)' : {
    #         #     'formula': 'C(4)H(5)N(1)O(1)', #84.0443902518
    #         #     'hcd':True
    #         # },
    #     },
    # },
    # "3,2′-O-dimethyluridine": {
    #     # m3Um
    #     "comments" : "Similarity to m3U assumed, Verfication required",
    #     "references" : ["Jora et al. 2018"],
    #     "precursors": {
    #         "3,2′-O-dimethyluridine": {
    #             "formula": "C(11)H(16)N(2)O(6)" # 273.1081127100
    #         }
    #     },
    #     "fragments": {
    #         "3,2′-O-dimethyluridine -Ribose": {
    #             "formula": "C(6)H(8)N(2)O(2)",  # 141.06585397237
    #             "hcd": True,
    #         },
    #         "3,2′-O-dimethyluridine -Ribose -C(1)H(5)N(1)": {
    #             "formula": "C(4)H(1)N(1)O(2)",  # 96.0080047430
    #             "hcd": True,
    #         },
    #     },
    # },
    "5-hydroxymethylcytidine": {
        "comments": "Elutes with m3C. Verification required. You et al. report peak 124.1",
        "references": ["You et al. 2019"],
        "precursors": {"5-hydroxymethylcytidine": {"formula": ""}},
        "fragments": {
            "5-hydroxymethylcytidine -Ribose": {
                "formula": "C(5)H(7)N(3)O(2)",  # 142.0611029442
                "hcd": False,
            },
            "5-hydroxymethylcytidine -Ribose -O(1)": {
                "formula": "C(5)H(7)N(3)O(1)",  #  126.0661883242
                "hcd": True,
            },
            "5-hydroxymethylcytidine -Ribose -H(3)N(1)O(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
                "hcd": True,
            },
            "5-hydroxymethylcytidine -Ribose -C(1)H(5)N(1)O(1)": {
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            },
        },
    },
    "N4,2′-O-dimethylcytidine": {
        # m4Cm
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {"N4,2′-O-dimethylcytidine": {}},
        "fragments": {
            "N4,2′-O-dimethylcytidine -Methylated ribose": {
                "formula": "C(5)H(7)N(3)O(1)"  # 126.0661883242
            },
            "N4,2′-O-dimethylcytidine -Methylated ribose -H(3)N(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
                "hcd": True,
            },
            "N4,2′-O-dimethylcytidine -Methylated ribose -H(2)O(1)": {
                "formula": "C(5)H(5)N(3)",  # 108.0556236398
                "hcd": True,
            },
            "N4,2′-O-dimethylcytidine -Methylated ribose -C(1)H(5)N(1)": {
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            },
            "N4,2′-O-dimethylcytidine -Methylated ribose -C(1)H(1)N(1)O(1)": {
                "formula": "C(4)H(6)N(2)",  # 83.0603746680
                "hcd": True,
            },
        },
    },
    "5,2′-O-dimethylcytidine": {
        # m5Cm
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {"5,2′-O-dimethylcytidine": {}},
        "fragments": {
            "5,2′-O-dimethylcytidine -Methylated ribose": {
                "formula": "C(5)H(7)N(3)O(1)"  # 126.0661883242
            },
            "5,2′-O-dimethylcytidine -Methylated ribose -H(3)N(1)": {
                "formula": "C(5)H(4)N(2)O(1)",  # 109.0396392236
                "hcd": True,
            },
            "5,2′-O-dimethylcytidine -Methylated ribose -H(2)O(1)": {
                "formula": "C(5)H(5)N(3)",  # 108.0556236398
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "5,2′-O-dimethylcytidine -Methylated ribose -C(1)H(5)N(1)": {  #
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            }
        },
    },
    "N4,N4-dimethylcytidine": {
        # m4Cm
        "comments": "Verification required",
        "references": [],
        "precursors": {"N4,N4-dimethylcytidine": {}},
        "fragments": {
            "N4,N4-dimethylcytidine -Ribose": {
                "formula": "C(6)H(9)N(3)O(1)"  # 140.0818383886
            }
        },
    },
    "1-methylguanosine": {
        # m1G
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {"1-methylguanosine": {}, "1-methylguanosine -Ribose": {}},
        "fragments": {
            "1-methylguanosine -Ribose": {
                "formula": "C(6)H(7)N(5)O(1)"  # 166.0723363322
            },
            "1-methylguanosine -Ribose -H(1)N(1) +O(1)": {
                "formula": "C(6)H(6)N(4)O(2)",  # 167.0563519160
                "hcd": True,
            },
            "1-methylguanosine -Ribose -C(1)H(3)N(1) +O(1)": {
                "formula": "C(5)H(4)N(4)O(2)",  # 153.0407018516
                "hcd": True,
            },
            "1-methylguanosine -Ribose -H(3)N(1)": {
                "formula": "C(6)H(4)N(4)O(1)",  # 149.0457872316
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "7-O-methylguanosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)O(1)"  # 152.0566862678
            }
        },
    },
    "N2-methylguanosine": {
        # m2G
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {"N2-methylguanosine": {}, "N2-methylguanosine -Ribose": {}},
        "fragments": {
            "N2-methylguanosine -Ribose": {
                "formula": "C(6)H(7)N(5)O(1)",  # 166.0723363322
                "hcd": False,
            },
            "N2-methylguanosine -Ribose -H(1)N(1) +O(1)": {
                "formula": "C(6)H(6)N(4)O(2)",  # 167.0563519160
                "hcd": True,
            },
            "N2-methylguanosine -Ribose -H(3)N(1)": {
                "formula": "C(6)H(4)N(4)O(1)",  # 149.0457872316
                "hcd": True,
            },
            "N2-methylguanosine -Ribose -C(2)H(2)N(2) +O(1)": {
                "formula": "C(4)H(5)N(3)O(2)",  # 128.0454528798
                "hcd": True,
            },
            "N2-methylguanosine -Ribose -C(2)H(4)N(2)": {
                "formula": "C(4)H(3)N(3)O(1)",  # 110.0348881954
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "N2-methylguanosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)O(1)"  # 152.0566862678
            },
            "N2-methylguanosine -Ribose": {
                "formula": "C(6)H(7)N(5)O(1)",  # 166.0723363322
                "hcd": True,
            },
        },
    },
    "7-methylguanosine": {
        # m7G
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {"7-methylguanosine": {}, "7-methylguanosine -Ribose": {}},
        "fragments": {
            "7-methylguanosine -Ribose": {
                "formula": "C(6)H(7)N(5)O(1)"  # 166.0723363322
            },
            "7-methylguanosine -Ribose -H(1)N(1) +O(1)": {
                "formula": "C(6)H(6)N(4)O(2)",  # 167.0563519160
                "hcd": True,
            },
            "7-methylguanosine -Ribose -H(3)N(1)": {
                "formula": "C(6)H(4)N(4)O(1)",  # 149.0457872316
                "hcd": True,
            },
            # '7-methylguanosine -Ribose -C(1)N(2) +O(1)': { #too low abudant for 5% TIC threshold
            #     'formula': 'C(5)H(7)N(3)O(2)', # 142
            #     'hcd': True,
            # },
            "7-methylguanosine -Ribose -C(1)H(2)N(2)": {
                "formula": "C(5)H(5)N(3)O(1)",  # 124.0505382598
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "7-O-methylguanosine -Methylated ribose": {"formula": "C(5)H(5)N(5)O(1)"}
        },
    },
    "2′-O-methylguanosine": {
        # Cm
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "2′-O-methylguanosine": {},
            "2′-O-methylguanosine -Methylated ribose": {},
            "2′-O-methylguanosine K adduct": {},
            "2′-O-methylguanosine dimer": {},
            "2′-O-methylguanosine K adduct dimer": {},
        },
        "fragments": {
            "2′-O-methylguanosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)O(1)"  # 152.0566862678
            },
            "2′-O-methylguanosine -Methylated ribose -H(1)N(1) +O(1)": {
                "formula": "C(5)H(4)N(4)O(2)",  # 153.0407018516
                "hcd": True,
            },
            "2′-O-methylguanosine -Methylated ribose -H(3)N(1)": {
                "formula": "C(5)H(2)N(4)O(1)",  # 135.0301371672
                "hcd": True,
            },
            "2′-O-methylguanosine -Methylated ribose -C(1)H(2)N(2)": {
                "formula": "C(4)H(3)N(3)O(1)",  # 110.0348881954
                "hcd": True,
            },
            # '7-methylguanosine -Ribose -C(1)N(2) +O(1)': { #too low abudant for 5% TIC threshold
            #     'formula': 'C(5)H(7)N(3)O(2)', # 142
            #     'hcd': True,
            # },
        },
        "exclusion_fragments": {
            "2′-O-methylguanosine -Ribose": {
                "formula": "C(6)H(7)N(5)O(1)",  # 166.0723363322
            }
        },
    },
    "N2,N2-dimethylguanosine": {
        # m2,2G
        "comments": "",
        "references": ["Giessing et al. 2011"],
        "precursors": {
            "N2,N2-dimethylguanosine": {},
            "N2,N2-dimethylguanosine dimer": {},
        },
        "fragments": {
            "N2,N2-dimethylguanosine -Ribose": {
                "formula": "C(7)H(9)N(5)O(1)"  # 180.08798639657002
            },
            "N2,N2-dimethylguanosine -Ribose -C(2)H(5)N(1) +O(1)": {
                "formula": "C(5)H(4)N(4)O(2)",  # 153.0407018516
                "hcd": True,
            },
            "N2,N2-dimethylguanosine -Ribose -C(2)H(7)N(1)": {
                "formula": "C(5)H(2)N(4)O(1)",  # 135.0301371672
                "hcd": True,
            },
            "N2,N2-dimethylguanosine -Ribose -C(3)H(6)N(2)": {
                "formula": "C(4)H(3)N(3)O(1)",  # 110.0348881954
                "hcd": True,
            },
        },
    },
    # "7-aminomethyl-7-deazaguanosine": {
    #     # is not validated!
    #     "precursors": {"7-aminomethyl-7-deazaguanosine": {}},
    #     "fragments": {
    #         "7-aminomethyl-7-deazaguanosine -Ribose": {
    #             "formula": "C(7)H(9)N(5)O(1)"  # 180.0879863966
    #         },
    #         "7-aminomethyl-7-deazaguanosine -Ribose -H(3)N(1)": {
    #             "formula": "C(7)H(6)N(4)O(1)",  # 163.0614372960
    #             "hcd": True,
    #         },
    #     },
    # },
    "adenosine": {
        # A
        "comments": "",
        "references": [],
        "precursors": {"adenosine": {}},
        "fragments": {
            "adenosine -Ribose": {"formula": "C(5)H(5)N(5)"},  # 136.0617716478
            "adenosine -Ribose -H(3)N(1)": {
                "formula": "C(5)H(2)N(4)",  # 119.03522254717
                "hcd": True,
            },
        },
    },
    "2′-O-methyladenosine": {
        # Am
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "2′-O-methyladenosine": {"formula": "C(11)H(15)N(5)O(4)"},  # 282.1196804498
            "2′-O-methyladenosine dimer": {},
        },
        "fragments": {
            "2′-O-methyladenosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)"  # 136.0617716478
            },
            # '2′-O-methyladenosine': {'formula': 'C(11)H(15)N(5)O(4)'}
        },
        "exclusion_fragments": {
            "2′-O-methyladenosine -Ribose": {
                "formula": "C(6)H(7)N(5)"  # 150.0774217122
            }
        },
    },
    "1-methyladenosine": {
        # m1A
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "1-methyladenosine": {"formula": "C(11)H(15)N(5)O(4)"},
            "1-methyladenosine dimer": {},
            "1-methyladenosine K adduct dimer": {},
        },
        "fragments": {
            "1-methyladenosine -Ribose": {"formula": "C(6)H(7)N(5)"},  # 150.0774217122
            "1-methyladenosine -Ribose -N(1)H(3)": {
                "formula": "C(6)H(4)N(4)",  # 133.0508726116
                "hcd": True,
            },
            "1-methyladenosine -Ribose -C(2)N(1)H(3)": {
                "formula": "C(4)H(4)N(4)",  # 109.0508726116
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "1-methyladenosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)"  # 136.0617716478
            }
        },
    },
    "1,2′-O-dimethyladenosine": {
        # m1Am
        "comments": "",
        "references": [],
        "precursors": {"1,2′-O-dimethyladenosine": {"formula": "C(12)H(17)N(5)O(4)"}},
        "fragments": {
            "1,2′-O-dimethyladenosine -Methylated ribose": {
                "formula": "C(6)H(7)N(5)"  # 150.0774217122
            },
            "1,2′-O-dimethyladenosine -Methylated ribose -N(1)H(3)": {
                "formula": "C(6)H(4)N(4)",  # 133.0508726116
                "hcd": True,
            },
            "1,2′-O-dimethyladenosine -Methylated ribose -C(2)N(1)H(3)": {
                "formula": "C(4)H(4)N(4)",  # 109.0508726116
                "hcd": True,
            },
        },
    },
    "N6,N6-dimethyladenosine": {
        # m6,6A
        "comments": "",
        "references": ["Chan et al. 2011"],
        "precursors": {
            "N6,N6-dimethyladenosine": {
                "formula": "C(12)H(17)N(5)O(4)"  # 296.1353305142
            }
        },
        "fragments": {
            "N6,N6-dimethyladenosine -Ribose": {
                "formula": "C(7)H(9)N(5)"  # 164.0930717766
            },
            "N6,N6-dimethyladenosine -Ribose -C(1)H(3)": {
                "formula": "C(6)H(6)N(5)",  # 149.06959667997
                "hcd": True,
            },
            "N6,N6-dimethyladenosine -Ribose -C(2)H(5)N(1)": {
                "formula": "C(5)H(4)N(4)",  # 121.05087261157
                "hcd": True,
            },
        },
    },
    "N6,2′-O-dimethyladenosine": {
        # m6Am
        "comments": "",
        "references": [],
        "precursors": {
            "N6,2′-O-dimethyladenosine": {
                "formula": "C(12)H(17)N(5)O(4)"  # 296.1353305142
            }
        },
        "fragments": {
            "N6,2′-O-dimethyladenosine -Methylated ribose": {
                "formula": "C(6)H(7)N(5)"  # 150.0774217122
            },
            "N6,2′-O-dimethyladenosine -Methylated ribose -C(1)H(1)N(1)": {
                "formula": "C(5)H(6)N(4)",  # 123.0665226760
                "hcd": True,
            },
            "N6,2′-O-dimethyladenosine -Methylated ribose -C(2)H(4)N(2)": {
                "formula": "C(4)H(3)N(3)",  # 94.0399735754
                "hcd": True,
            },
        },
    },
    "2,8-dimethyladenosine": {
        # m2,8A
        "comments": "Verification required",
        "references": ["Giessing et al. 2009"],
        "precursors": {"2,8-dimethyladenosine": {}},
        "fragments": {
            "2,8-dimethyladenosine -Ribose": {
                "formula": "C(7)H(9)N(5)"  # 164.0930717766
            },
            "2,8-dimethyladenosine -Ribose -H(3)N(1)": {
                "formula": "C(7)H(6)N(4)",  # 147.0665226760
                "hcd": True,
            },
        },
    },
    "2-methyladenosine": {
        # m2A
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "2-methyladenosine": {},
            "2-methyladenosine dimer": {},
            "2-methyladenosine K adduct dimer": {},
        },
        "fragments": {
            "2-methyladenosine -Ribose": {"formula": "C(6)H(7)N(5)"},  # 150.0774217122
            "2-methyladenosine -Ribose -N(1)H(3)": {
                "formula": "C(6)H(4)N(4)",  # 133.0508726116
                "hcd": True,
            },
            "2-methyladenosine -Ribose -C(2)N(2)H(6)": {
                "formula": "C(4)H(1)N(3)",  # 92.0243235110
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "2-methyladenosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)"  # 136.0617716478
            }
        },
    },
    "8-methyladenosine": {
        # m8A
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "8-methyladenosine": {},
            "8-methyladenosine dimer": {},
            "8-methyladenosine K adduct dimer": {},
        },
        "fragments": {
            "8-methyladenosine -Ribose": {"formula": "C(6)H(7)N(5)"},  # 150.0774217122
            "8-methyladenosine -Ribose -N(1)H(3)": {
                "formula": "C(6)H(4)N(4)",  # 133.0508726116
                "hcd": True,
            },
            "8-methyladenosine -Ribose -C(1)N(2)H(2)": {
                "formula": "C(5)H(5)N(3)",  # 108.0556236398
                "hcd": True,
            },
            "8-methyladenosine -Ribose -C(1)N(2)H(4)": {
                "formula": "C(5)H(3)N(3)",  # 106.0399735754
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "8-methyladenosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)"  # 136.0617716478
            }
        },
    },
    "N6-methyladenosine": {
        # m6A
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "N6-methyladenosine": {},
            "N6-methyladenosine dimer": {},
            "N6-methyladenosine K adduct dimer": {},
        },
        "fragments": {
            "N6-methyladenosine -Ribose": {"formula": "C(6)H(7)N(5)"},  # 150.0774217122
            "N6-methyladenosine -Ribose -C(1)H(1)N(1)": {
                "formula": "C(5)H(6)N(4)",  # 123.0665226760
                "hcd": True,
            },
            "N6-methyladenosine -Ribose -C(2)H(4)N(2)": {
                "formula": "C(4)H(3)N(3)",  # 94.0399735754
                "hcd": True,
            },
        },
        "exclusion_fragments": {
            "N6-methyladenosine -Methylated ribose": {
                "formula": "C(5)H(5)N(5)"  # 136.0617716478
            }
        },
    },
    # 'N6-acetyladenosine': {
    #      "references":['Sauerwald et al. 2005'],
    #     'precursors':{
    #         'N6-acetyladenosine': {'formula': 'C(12)H(15)N(5)O(5)'}
    #     },
    #     'fragments' : {
    #         'N6-acetyladenosine -Ribose': {
    #             'formula': 'C(7)H(7)N(5)O(1)',
    #             'hcd' : False
    #         },
    #         'N6-acetyladenosine -Ribose -C(1)O(1)': {
    #             'formula': 'C(6)H(7)N(5)',
    #             'hcd' : True
    #         },
    #         'N6-acetyladenosine -Ribose -C(2)H(1)N(1)O(1)': {
    #             'formula': 'C(5)H(6)N(4)', #123.06
    #             'hcd' : True
    #         },
    #         'N6-acetyladenosine -Ribose -C(3)H(4)N(2)O(1)': {
    #             'formula': 'C(4)H(3)N(3)', #94.04
    #             'hcd' : True
    #         },
    #     },
    # },
    "N6-hydroxymethyladenosine": {
        # hm6A
        "comments": "Verification required",
        "references": ["modomics.org", "You et al. 2019", "Fu et al. 2013"],
        "fragments": {
            "N6-hydroxymethyladenosine": {
                "formula": "C(11)H(15)N(5)O(5)"  # 298.1145950698
            },
            "adenosine": {"formula": "C(10)H(13)N(5)O(4)"},  # 268.1040303854
            "adenosine -Ribose": {"formula": "C(5)H(5)N(5)"},  # 136.0617716478
        },
        "precursors": {"N6-hydroxymethyladenosine": {}},
    },
    "N4-acetylcytidine": {
        # ac4C
        "comments": "Verification required",
        "references": ["modomics.org"],
        "fragments": {
            "N4-acetylcytidine -Ribose": {
                "formula": "C(6)H(7)N(3)O(2)",  # 154.0611029442
                "hcd": False,
            },
            "cytidine -Ribose": {"formula": "C(4)H(5)N(3)O(1)"},  # 112.0505382598
            "cytidine -Ribose -C(1)H(5)N(1)": {
                "formula": "C(4)H(2)N(2)O(1)",  # 95.0239891592
                "hcd": True,
            },
        },
        "precursors": {"N4-acetylcytidine": {}, "N4-acetylcytidine K adduct": {}},
    },
    "5-formyl-2′-O-methylcytidine": {
        # f5Cm
        "comments": "Verification required",
        "references": ["modomics.org"],
        "fragments": {
            "5-formyl-2′-O-methylcytidine -Methylated ribose": {
                "formula": "C(5)H(5)N(3)O(2)"  # 140.0454528798
            }
        },
        "precursors": {"5-formyl-2′-O-methylcytidine": {}},
    },
    "4-thiouridine": {
        # s4U
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {"4-thiouridine": {}},
        "fragments": {
            "4-thiouridine -Ribose": {
                "formula": "C(4)H(4)N(2)O(1)S(1)"  # 129.0117103976
            },
            "4-thiouridine -Ribose -H(3)N(1)": {
                "formula": "C(4)H(1)N(1)O(1)S(1)",  #  111.9851612970
                "hcd": True,
            },
            "4-thiouridine -Ribose -C(1)H(1)N(1)O(1)": {
                "formula": "C(3)H(3)N(1)S(1)",  # 86.0058967414
                "hcd": True,
            },
        },
    },
    "2-thiouridine": {
        # s2U
        "comments": "",
        "references": ["Jora et al. 2018"],
        "precursors": {
            "2-thiouridine": {},
            "2-thiouridine isotope_1": {
                "formula": "+C(9)H(12)N(2)O(5)S(1)",
                "isotope_position": 1,
            },  # refers to pos 1 in most abundant precursors
            # default would be 0, following the ursgal nomenclature, see below
        },
        "fragments": {
            "2-thiouridine -Ribose": {
                "formula": "C(4)H(4)N(2)O(1)S(1)"  # 129.0117103976
            },
            "2-thiouridine -Ribose -H(3)N(1)": {
                "formula": "C(4)H(1)N(1)O(1)S(1)",  #  111.9851612970
                "hcd": True,
            },
            "2-thiouridine -Ribose -C(1)H(1)N(1)S(1)": {
                "formula": "C(3)H(3)N(1)O(1)",  #  70.0287401874
                "hcd": True,
            },
        },
    },
    "spiroiminodihydantoin": {
        "comments": "",
        "references": ["Martinez et al. 2007"],
        "precursors": {
            "spiroiminodihydantoin": {"formula": "C(10)H(13)N(5)O(7)"}  # 316.0887742454
        },
        "fragments": {
            "spiroiminodihydantoin -Ribose": {
                "formula": "C(5)H(5)N(5)O(3)",  # 184.0465155078
                "hcd": False,
            },
            "spiroiminodihydantoin -Ribose -C(2)H(2)N(2)O(1)": {
                "formula": "C(3)H(3)N(3)O(2)",  # 114.0298028154
                "hcd": True,
            },
            "spiroiminodihydantoin -Ribose -C(2)H(1)N(1)O(2)": {
                "formula": "C(3)H(4)N(4)O(1)",  # 113.0457872316
                "hcd": True,
            },
            "spiroiminodihydantoin -Ribose -C(3)H(2)N(2)O(2)": {
                "formula": "C(2)H(3)N(3)O(1)",  # 86.0348881954
                "hcd": True,
            },
        },
    },
    "queuosine": {
        # Q
        "comments": "",
        "references": ["Zallot et al. 2014", "Giessing et al. 2009"],
        "precursors": {"queuosine": {"formula": "C(17)H(23)N(5)O(7)"}},
        "fragments": {
            "queuosine -C(5)H(9)N(1)O(2)": {
                "formula": "C(12)H(14)N(4)O(5)",  # 295.10369603357
                "hcd": False,
            },
            "queuosine -C(10)H(17)N(1)O(6)": {
                "formula": "C(7)H(6)N(4)O(1)",  # 163.0614372959700
                # 'hcd' : False
            },
            "queuosine -C(10)H(20)N(2)O(6)": {
                "formula": "C(7)H(3)N(3)O(1)",  # 146.0348881954
                "hcd": True,
            },
            "queuosine -C(11)H(19)N(3)O(6)": {
                "formula": "C(6)H(4)N(2)O(1)",  # 121.0396392236
                "hcd": True,
            },
        },
    },
}



"""Summary.
"""
import csv
import os
from tempfile import NamedTemporaryFile

import pytest

from smiter.lib import (
    check_mzml_params,
    check_peak_properties,
    csv_to_peak_properties,
    peak_properties_to_csv,
)


def test_check_mzml_params():
    mzml_params = {
        "gradient_length": 45,
    }
    mzml_params = check_mzml_params(mzml_params)
    assert mzml_params["gradient_length"] == 45


def test_check_faulty_mzml_params():
    mzml_params = {}
    with pytest.raises(Exception):
        mzml_params = check_mzml_params(mzml_params)


def test_check_peak_properties_missing_default():
    peak_props = {
        "ELVISLIVES": {
            "trivial_name": "ELVISLIVES",
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width
        },
    }
    peak_props = check_peak_properties(peak_props)
    assert peak_props["ELVISLIVES"]["charge"] == 2


def test_check_peak_properties_missing_opt():
    peak_props = {
        "ELVISLIVES": {
            "scan_start_time": 0,
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width
        },
    }
    peak_props = check_peak_properties(peak_props)
    # sets charge but leaves out trivial_name
    assert peak_props["ELVISLIVES"]["charge"] == 2
    assert "trivial_name" not in peak_props["ELVISLIVES"]


def test_check_peak_properties_missing_required():
    peak_props = {
        "ELVISLIVES": {
            "trivial_name": "ELVISLIVES",
            "peak_width": 30,  # seconds
            "peak_function": "gauss",
            "peak_params": {"sigma": 3},  # 10% of peak width
        },
    }
    with pytest.raises(Exception):
        peak_props = check_peak_properties(peak_props)


def test_csv_to_peak_properties():
    csv_file = os.path.join(os.path.dirname(__file__), "data", "molecules_test.csv")
    peak_properties = csv_to_peak_properties(csv_file)
    assert "2′-O-methylcytidine" in peak_properties
    assert (
        peak_properties["2′-O-methylcytidine"]["trivial_name"] == "2′-O-methylcytidine"
    )
    assert peak_properties["2′-O-methylcytidine"]["scan_start_time"] == 10
    assert peak_properties["2′-O-methylcytidine"]["peak_scaling_factor"] == 1e6
    # default
    assert peak_properties["2′-O-methylcytidine"]["charge"] == 2

    assert "5-methoxycarbonylmethyluridine" in peak_properties
    assert (
        peak_properties["5-methoxycarbonylmethyluridine"]["trivial_name"]
        == "5-methoxycarbonylmethyluridine"
    )
    assert peak_properties["5-methoxycarbonylmethyluridine"]["scan_start_time"] == 12
    assert (
        peak_properties["5-methoxycarbonylmethyluridine"]["peak_scaling_factor"] == 2e6
    )
    # default
    assert peak_properties["5-methoxycarbonylmethyluridine"]["charge"] == 2


def test_peak_properties_to_csv():
    peak_properties = {
        "2′-O-methylcytidine": {
            "trivial_name": "2′-O-methylcytidine",
            "chemical_formula": "+C(10)H(15)N(3)O(5)",
            "charge": 2,
            "scan_start_time": 10.0,
            "peak_function": "gauss",
            "peak_params": {"sigma": 2},
            "peak_scaling_factor": 1000000.0,
            "peak_width": 30,
        },
        "5-methoxycarbonylmethyluridine": {
            "trivial_name": "5-methoxycarbonylmethyluridine",
            "chemical_formula": "+C(12)H(16)N(2)O(8)",
            "charge": 2,
            "scan_start_time": 12.0,
            "peak_function": "gauss",
            "peak_params": {"sigma": 2},
            "peak_scaling_factor": 2000000.0,
            "peak_width": 30,
        },
    }
    csv_file = "out.csv"
    fname = peak_properties_to_csv(peak_properties, csv_file)
    assert fname == csv_file
    with open(csv_file) as fout:
        reader = csv.DictReader(fout)
        lines = [l for l in reader]
    print(lines)

    assert lines[0]["chemical_formula"] == "+C(10)H(15)N(3)O(5)"
    assert lines[0]["scan_start_time"] == "10.0"
    assert lines[0]["peak_function"] == "gauss"
    assert lines[0]["peak_params"] == "sigma=2"
    assert lines[0]["peak_scaling_factor"] == "1000000.0"
    assert lines[0]["peak_width"] == "30"
    # default
    assert lines[0]["charge"] == "2"

    assert lines[1]["chemical_formula"] == "+C(12)H(16)N(2)O(8)"
    assert lines[1]["scan_start_time"] == "12.0"
    assert lines[1]["peak_function"] == "gauss"
    assert lines[1]["peak_params"] == "sigma=2"
    assert lines[1]["peak_scaling_factor"] == "2000000.0"
    assert lines[1]["peak_width"] == "30"
    # default
    assert lines[0]["charge"] == "2"



"""Callables for injection noise into scans.

Upon calling the callabe, a list/np.array of mz and intensities should be returned.
Arguments should be passed via *args and **kwargs
"""
import math
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

import numpy as np
import pyqms
from loguru import logger

import smiter
from smiter.lib import calc_mz

# from smiter.synthetic_mzml import Scan
# import smiter.synthetic_mzml


class AbstractNoiseInjector(ABC):
    """Summary."""

    def __init__(self, *args, **kwargs):
        """Initialize noise injector."""
        pass  # pragma: no cover

    @abstractmethod
    def inject_noise(self, scan, *args, **kwargs):
        """Main noise injection method.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description

        """
        pass  # pragma: no cover

    @abstractmethod
    def _ms1_noise(self, scan, *args, **kwargs):
        pass  # pragma: no cover

    @abstractmethod
    def _msn_noise(self, scan, *args, **kwargs):
        pass  # pragma: no cover


class GaussNoiseInjector(AbstractNoiseInjector):
    def __init__(self, *args, **kwargs):
        # np.random.seed(1312)
        logger.info("Initialize GaussNoiseInjector")
        self.args = args
        self.kwargs = kwargs

    def inject_noise(self, scan, *args, **kwargs):
        """Main noise injection method.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description

        """
        self.kwargs.update(kwargs)
        # self.args += args
        if scan.ms_level == 1:
            scan = self._ms1_noise(scan, *self.args, **self.kwargs)
        elif scan.ms_level > 1:
            scan = self._msn_noise(scan, *args, **kwargs)
        return scan

    def _ms1_noise(self, scan, *args, **kwargs):
        """Generate ms1 noise.

        Args:
            scan (Scan): Description
            *args: Description
            **kwargs: Description
        """
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise
        scan.i[scan.i < 0] = 0
        return scan

    def _msn_noise(self, scan, *args, **kwargs):
        """Generate msn noise.

        Args:
            scan (Scan): Description
            *args: Description
            **kwargs: Description
        """
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise
        scan.i[scan.i < 0] = 0
        # remove some mz values
        dropout = kwargs.get("dropout", 0.1)
        dropout_mask = [
            False if c < dropout else True for c in np.random.random(len(scan.mz))
        ]
        scan.mz = scan.mz[dropout_mask]
        scan.i = scan.i[dropout_mask]
        return scan

    def _generate_mz_noise(self, scan, *args, **kwargs):
        """Generate noise for mz_array.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description
        """
        ppm_offset = kwargs.get("ppm_offset", 0)
        ppm_var = kwargs.get("ppm_var", 1)
        noise_level = np.random.normal(0, ppm_var * 1e-6, len(scan.mz))
        # noise_level = np.zeros(len(scan.mz))
        noise = scan.mz * noise_level
        # noise = np.zeros(len(scan.mz))
        return noise

    def _generate_intensity_noise(self, scan, *args, **kwargs):
        """Generate intensity noise.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description
        """
        noise = np.random.normal(
            # np.array(np.zeros(len(scan.i))),
            np.array(scan.i),
            np.array(scan.i) * kwargs.get("variance", 0.02),
            len(scan.i),
        )
        return noise - scan.i


class UniformNoiseInjector(AbstractNoiseInjector):
    def __init__(self, *args, **kwargs):
        # np.random.seed(1312)
        logger.info("Initialize UniformNoiseInjector")
        self.args = args
        self.kwargs = kwargs

    def inject_noise(self, scan, *args, **kwargs):
        """Main noise injection method.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description

        """
        self.kwargs.update(kwargs)
        # self.args += args
        if scan.ms_level == 1:
            scan = self._ms1_noise(scan, *self.args, **self.kwargs)
        elif scan.ms_level > 1:
            scan = self._msn_noise(scan, *self.args, **self.kwargs)
        return scan

    def _ms1_noise(self, scan, *args, **kwargs):
        """Generate ms1 noise.

        Args:
            scan (Scan): Description
            *args: Description
            **kwargs: Description
        """
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise
        scan.i[scan.i < 0] = 0
        return scan

    def _msn_noise(self, scan, *args, **kwargs):
        """Generate msn noise.

        Args:
            scan (Scan): Description
            *args: Description
            **kwargs: Description
        """
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise
        scan.i[scan.i < 0] = 0
        # scan.mz[scan.mz < 0] = 0
        dropout = kwargs.get("dropout", 0.1)
        dropout_mask = [
            False if c < dropout else True for c in np.random.random(len(scan.mz))
        ]
        scan.mz = scan.mz[dropout_mask]
        scan.i = scan.i[dropout_mask]
        return scan

    def _generate_mz_noise(self, scan, *args, **kwargs):
        """Generate noise for mz_array.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description
        """
        # noise_level = np.random.normal(ppm_offset * 5e-6, ppm_var * 5e-6, len(scan.mz))
        # get scaling from kwargs
        noise = np.random.uniform(
            (scan.mz * kwargs.get("ppm_noise", 5e-6)) * -1,
            scan.mz * kwargs.get("ppm_noise", 5e-6),
        )
        # noise = scan.mz * noise_level
        return noise

    def _generate_intensity_noise(self, scan, *args, **kwargs):
        """Generate intensity noise.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description
        """
        # get scaling from kwargs
        noise = np.random.uniform(
            (scan.i * kwargs.get("intensity_noise", 0.2)) * -1,
            scan.i * kwargs.get("intensity_noise", 0.2),
        )
        return noise


class PPMShiftInjector(AbstractNoiseInjector):
    def __init__(self, *args, **kwargs):
        # np.random.seed(1312)
        logger.info("Initialize PPMShiftInjector")
        self.args = args
        self.kwargs = kwargs

    def inject_noise(self, scan, *args, **kwargs):
        """Main noise injection method.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description

        """
        self.kwargs.update(kwargs)
        # self.args += args
        if scan.ms_level == 1:
            scan = self._ms1_noise(scan, *self.args, **self.kwargs)
        elif scan.ms_level > 1:
            scan = self._msn_noise(scan, *self.args, **self.kwargs)
        return scan

    def _ms1_noise(self, scan, *args, **kwargs):
        """Generate ms1 noise.

        Args:
            scan (Scan): Description
            *args: Description
            **kwargs: Description
        """
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise
        scan.i[scan.i < 0] = 0
        return scan

    def _msn_noise(self, scan, *args, **kwargs):
        """Generate msn noise.

        Args:
            scan (Scan): Description
            *args: Description
            **kwargs: Description
        """
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise
        scan.i[scan.i < 0] = 0
        # scan.mz[scan.mz < 0] = 0
        dropout = kwargs.get("dropout", 0.1)
        dropout_mask = [
            False if c < dropout else True for c in np.random.random(len(scan.mz))
        ]
        scan.mz = scan.mz[dropout_mask]
        scan.i = scan.i[dropout_mask]
        return scan

    def _generate_mz_noise(self, scan, *args, **kwargs):
        """Generate noise for mz_array.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description
        """
        # noise_level = np.random.normal(ppm_offset * 5e-6, ppm_var * 5e-6, len(scan.mz))
        # get scaling from kwargs
        offset = self.kwargs["offset"]
        sigma = self.kwargs["sigma"]
        noise = np.random.normal(offset, sigma, len(scan.mz))
        noise = noise * scan.mz
        return noise

    def _generate_intensity_noise(self, scan, *args, **kwargs):
        """Generate intensity noise.

        Args:
            scan (Scan): Scan object
            *args: Description
            **kwargs: Description
        """
        # get scaling from kwargs
        noise = [0 for x in range(len(scan.i))]
        return noise


# TODO add white noise params
# TODO should be called MSpireNoiseInjector
class JamssNoiseInjector(AbstractNoiseInjector):
    def __init__(self, *args, **kwargs):
        """Noise injector based on this paper:
            https://academic.oup.com/bioinformatics/article/31/5/791/318378

        Args:
            **kwargs: fine tuning for mz and itensity sigma
                - m: scale mz noise sigma by this value
                - y: controll falling of mz sigma with higher intensities with this
                - a: scale intensity sigma by this
                - b: controll falling of intensity sigma with higher intensities with this
                - c: constant to add to intensity sigma
        """
        # np.random.seed(1312)
        logger.info("Initialize JamssNoiseInjector")
        self.args = args
        self.kwargs = kwargs

    def inject_noise(self, scan, *args, **kwargs):
        self.kwargs.update(kwargs)
        if scan.ms_level == 1:
            scan = self._ms1_noise(scan, *self.args, **self.kwargs)
        elif scan.ms_level > 1:
            scan = self._msn_noise(scan, *self.args, **self.kwargs)
        return scan

    def _add_white_noise(self, scan):
        n = int(np.random.uniform(100, 500))
        if sum(scan.i) == 0:
            total_tic = 5e6
        else:
            total_tic = sum(scan.i)
            total_tic = 5e6
        # logger.info(f"Total TIC: {sum(scan.i)}")
        white_noise_i = np.random.uniform(1, 100, n)
        max_perc_noise = 0.75
        min_perc_noise = 0.5
        white_noise_i = (
            white_noise_i
            / white_noise_i.sum()
            * np.random.uniform(min_perc_noise, max_perc_noise)
            * total_tic
        )
        # logger.info(f'Add white noise intensity: {white_noise_i}')
        # scale = np.random.uniform(1e5, 1e8) / n
        # white_noise_i *= scale
        # logger.info(f"Noise: {sum(white_noise_i)}")
        white_noise_mz = np.random.uniform(0, 1200, n)
        new_i = np.concatenate((scan.i, white_noise_i))
        new_mz = np.concatenate((scan.mz, white_noise_mz))
        # logger.info(f'Add new white noise {new_i}')
        sort = new_mz.argsort()
        scan.mz = new_mz[sort]
        scan.i = new_i[sort]
        return scan

    def _ms1_noise(self, scan, *args, **kwargs):
        # i_b = sum(scan.i)
        scan = self._add_white_noise(scan)
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        # logger.debug(f"MS1 mz noise: {mz_noise}")
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise

        # i_a = sum(scan.i)
        # print(f'S2I: {i_b/i_a}')
        scan.i[scan.i < 0] = 0
        return scan

    def _msn_noise(self, scan, *args, **kwargs):
        mz_noise = self._generate_mz_noise(scan, *args, **kwargs)
        # logger.debug(f"MSn mz noise: {mz_noise}")
        intensity_noise = self._generate_intensity_noise(scan, *args, **kwargs)
        scan.mz += mz_noise
        scan.i += intensity_noise
        scan.i[scan.i < 0] = 0
        # scan.mz[scan.mz < 0] = 0
        dropout = kwargs.get("dropout", 0.1)
        # logger.debug(f"Dropout: {dropout}")
        dropout_mask = [
            False if c < dropout else True for c in np.random.random(len(scan.mz))
        ]
        scan.mz = scan.mz[dropout_mask]
        scan.i = scan.i[dropout_mask]
        scan = self._add_white_noise(scan)
        return scan

    def _generate_mz_noise(self, scan, *args, **kwargs):
        try:
            norm_int = scan.i / max(scan.i) * 100
        except ValueError:
            norm_int = np.array([])
        # TODO make variable parameters
        # TODO Check mspire paper again for variables
        # TODO pass data structure with max_i in elution profile and mzs calculate noise to intensity/max_i_in_profile
        m = 0.001701
        # m = 0.1701
        y = 0.543
        y = 0.2
        sigma = m * norm_int ** (-y)
        noise = np.random.normal(loc=scan.mz, scale=sigma)
        noise -= scan.mz
        noise_mock = np.zeros(len(scan.mz))
        return noise

    def _generate_intensity_noise(self, scan, *args, **kwargs):
        try:
            norm_int = scan.i / max(scan.i) * 100
        except ValueError:
            norm_int = np.array([])
        # TODO make variable parameters
        # TODO Check mspire paper again for variables
        # TODO pass data structure with max_i in elution profile and mzs calculate noise to intensity/max_i_in_profile
        m = 0.001701
        m = 10.34
        c = 0.00712
        d = 0.12
        sigma = m * (1 - math.e ** (-c * norm_int)) + d
        noise = np.random.normal(loc=0, scale=sigma)
        if len(scan.i) > 0:
            noise *= max(scan.i) / 100
        else:
            noise = np.array([])
        # TODO add white noise points with noise = x * tic (0 < x < 1)
        return noise



#!/usr/bin/env python3
"""Distribution funtion for chromo peaks.

Attributes:
    distributions (dict): mapping distribution name to distribution function
"""
import math
from typing import Callable, Dict

from loguru import logger
from scipy.stats import gamma


def gauss_dist(x: float, sigma: float = 1, mu: float = 0):
    """Calc Gauss distribution.

    Args:
        x (float): x
        sigma (float, optional): standard deviation
        mu (float, optional): mean

    Returns:
        float: y
    """
    # logger.debug(f"Sigma: {sigma}")
    # print(f'Sigma: {sigma}')
    return (
        (
            1
            / (sigma * math.sqrt(2 * math.pi))
            * pow(math.e, (-0.5 * pow(((x - mu) / sigma), 2)))
        )
        / 0.3989422804014327
        * sigma
    )


def gauss_tail(
    x: float,
    mu: float,
    sigma: float,
    scan_start_time: float,
    h: float = 1,
    t: float = 0.2,
    f: float = 0.01,
) -> float:
    # h = 1
    ## should be a parameter
    # t = 0.2
    # f = 0.01
    sigma = t * (x - scan_start_time) + f  # / (x + 0.00001)
    i = h * math.e ** (-0.5 * ((x - mu) / sigma) ** 2)
    return i


def gamma_dist(x: float, a: float = 5, scale: float = 0.33):
    """Calc gamma distribution.

    Args:
        x (float): Description
        a (float, optional): Description
        scale (float, optional): Description

    Returns:
        float: y
    """
    return gamma.pdf(x, a=a, scale=scale)


distributions = {
    "gauss": gauss_dist,
    "gamma": gamma_dist,
    "gauss_tail": gauss_tail,
}  # type: Dict[str, Callable]



"""Callables for fragmenting molecules.

Upon calling the callabe, a list/np.array of mz and intensities should be returned.
Arguments should be passed via *args and **kwargs
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union
import sys
import shutil
import subprocess
import os
import csv

import numpy as np
import pandas as pd
import pyqms
from loguru import logger
from pyteomics import mass

import smiter
from peptide_fragmentor import PeptideFragment0r
from smiter.ext.nucleoside_fragment_kb import (
    KB_FRAGMENTATION_INFO as pyrnams_nucleoside_fragment_kb,
)
from smiter.lib import calc_mz

try:
    from smiter.ext.nucleoside_fragment_kb import KB_FRAGMENTATION_INFO
except ImportError:  # pragma: no cover
    print("Nucleoside fragmentation KB not available")  # pragma: no cover


class AbstractFragmentor(ABC):
    """Summary."""

    @abstractmethod
    def __init__(self):
        """Summary."""
        pass  # pragma: no cover

    @abstractmethod
    def fragment(self, entity):
        """Summary.

        Args:
            entity (TYPE): Description
        """
        pass  # pragma: no cover


class PeptideFragmentor(AbstractFragmentor):
    """Summary."""

    def __init__(self, *args, **kwargs):
        """Summary."""
        logger.info("Initialize PeptideFragmentor")
        self.args = args
        self.kwargs = kwargs
        self.fragger = PeptideFragment0r()

    # @profile
    def fragment(self, entities):
        """Summary.

        Args:
            entity (TYPE): Description
        """
        if isinstance(entities, str):
            entities = [entities]
        frames = []
        for entity in entities:
            # logger.debug(f"Fragment {entity}")
            results_table = self.fragger.fragment(entity, **self.kwargs)
            frames.append(results_table)
        final_table = pd.concat(frames)
        i = np.array([100 for i in range(len(final_table))])
        mz_i = np.stack((final_table["mz"], i), axis=1)
        return mz_i


class PeptideFragmentorPyteomics(AbstractFragmentor):
    def __init__(self, *args, **kwargs):
        pass

    # @profile
    def _fragments(self, peptide, types=("b", "y"), maxcharge=1):
        for i in range(1, len(peptide) - 1):
            for ion_type in types:
                for charge in range(1, maxcharge + 1):
                    if ion_type[0] in "abc":
                        yield mass.fast_mass(
                            peptide[:i], ion_type=ion_type, charge=charge
                        )
                    else:
                        yield mass.fast_mass(
                            peptide[i:], ion_type=ion_type, charge=charge
                        )

    # @profile
    def fragment(self, entities):
        mz = []
        for e in entities:
            a = list(self._fragments(e))
            mz.extend(a)
        i = np.array([100 for x in range(len(mz))])
        mz = np.array(mz)
        mz_i = np.stack((mz, i), axis=1)
        return mz_i


class NucleosideFragmentor(AbstractFragmentor):
    """Summary."""

    def __init__(
        self,
        nucleotide_fragment_kb: Dict[str, dict] = None,
        raise_error_for_non_existing_fragments=True,
    ):
        """Summary."""
        logger.info("Initialize NucleosideFragmentor")
        if nucleotide_fragment_kb is None:
            nucleoside_fragment_kb = pyrnams_nucleoside_fragment_kb
        self.raise_error_for_non_existing_fragments = (
            raise_error_for_non_existing_fragments
        )
        nuc_to_fragments: Dict[str, List[float]] = {}
        cc = pyqms.chemical_composition.ChemicalComposition()
        for nuc_name, nuc_dict in nucleoside_fragment_kb.items():
            nuc_to_fragments[nuc_name] = []
            for frag_name, frag_cc_dict in nucleoside_fragment_kb[nuc_name][
                "fragments"
            ].items():
                cc.use(f"+{frag_cc_dict['formula']}")
                m = cc._mass()
                nuc_to_fragments[nuc_name].append(calc_mz(m, 1))
        self.nuc_to_fragments = nuc_to_fragments

    def fragment(
        self, entities: Union[list, str], raise_error_for_non_existing_fragments=False
    ):
        """Summary.

        Args:
            entity (TYPE): Description
        """
        if isinstance(entities, str):
            entities = [entities]
        m = []
        for entity in entities:
            if raise_error_for_non_existing_fragments is True:
                masses = self.nuc_to_fragments[entity]
            else:
                masses = self.nuc_to_fragments.get(entity, [])
            m.extend(masses)
            # logger.debug(masses)
            # should overlapping peaks be divided into two very similar ones?
        m = sorted(list(set(m)))
        # logger.debug(m)
        return np.array([(mass, 1) for mass in m])


class LipidFragmentor(AbstractFragmentor):
    """Summary."""

    def __init__(
        self,
        lipid_input_csv: str = None,
        raise_error_for_non_existing_fragments=True,
    ):
        """Use LipidCreator to calculate precursor transitions of lipids."""
        self.lip_to_fragments = {}
        # TODO run lipid fragmenter here, read output file and collect results in dict
        commands: List[str] = []
        if sys.platform == "linux" or sys.platform == "darwin":
            commands.append("mono")
            lipid_creator_path = shutil.which("LipidCreator.exe")
        else:
            # will this work under windows?
            lipid_creator_path = "LipidCreator"
        commands.extend(
            [lipid_creator_path, "transitionlist", lipid_input_csv, "lipid_output.csv"]
        )
        proc = subprocess.run(commands)
        with open("lipid_output.csv") as fin:
            for line in csv.DictReader(fin):
                if line["PrecursorName"] not in self.lip_to_fragments:
                    self.lip_to_fragments[line["PrecursorName"]] = []
                self.lip_to_fragments[line["PrecursorName"]].append(
                    float(line["ProductMz"])
                )
        os.remove("lipid_output.csv")

    def fragment(
        self, entities: Union[list, str], raise_error_for_non_existing_fragments=False
    ):
        """Summary.

        Args:
            entity (TYPE): Description
        """
        if isinstance(entities, str):
            entities = [entities]
        m = []
        for entity in entities:
            if raise_error_for_non_existing_fragments is True:
                masses = self.lip_to_fragments[entity]
            else:
                masses = self.lip_to_fragments.get(entity, [])
            m.extend(masses)
            # logger.debug(masses)
            # should overlapping peaks be divided into two very similar ones?
        m = sorted(list(set(m)))
        # logger.debug(m)
        return np.array([(mass, 1) for mass in m])



"""Summary."""
import numpy as np
import pytest

import smiter
from smiter.noise_functions import (
    GaussNoiseInjector,
    UniformNoiseInjector,
    JamssNoiseInjector,
)
from smiter.synthetic_mzml import Scan

# np.random.seed(1312)

# make it a test class
# integration test, test GaussNoiseInjector methods by themselves!
def test_GaussNoiseInjector_intensity_noise_ms1():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = GaussNoiseInjector(variance=0.2)
    scan = noise_injector.inject_noise(scan)
    # breakpoint()
    # assert abs((scan.i - 3e6) < 3e6 * 0.05).all()
    assert (scan.i - i_array < i_array * 0.2).all()


# make it a test class
# integration test, test GaussNoiseInjector methods by themselves!
def test_GaussNoiseInjector_intensity_noise_ms2():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 2,
        }
    )
    noise_injector = GaussNoiseInjector()
    scan = noise_injector.inject_noise(scan, dropout=0.0)
    assert (scan.i - i_array < i_array * 0.2).all()


def test_GaussNoiseInjector_intensity_noise_ms2_drop_all():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 2,
        }
    )
    noise_injector = GaussNoiseInjector()
    scan = noise_injector.inject_noise(scan, dropout=1.0)
    assert len(scan.mz) == 0
    assert len(scan.i) == 0


def test_GaussNoiseInjector_generate_intensity_noise():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = GaussNoiseInjector()
    noise = noise_injector._generate_intensity_noise(scan)
    assert (abs(noise) < i_array * 0.2).all()


def test_GaussNoiseInjector_generate_mz_noise():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = GaussNoiseInjector()
    noise = noise_injector._generate_mz_noise(scan)
    # breakpoint()
    assert (abs(noise) < scan.mz * 1e-5).all()


def test_gauss_mz_noise_ms1():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = GaussNoiseInjector()
    scan = noise_injector.inject_noise(scan)
    # breakpoint()
    assert (
        abs(scan.mz - np.array([100, 200, 300], dtype="float32"))
        < np.array([100, 200, 300], dtype="float32") * 1e-5
    ).all()


def test_uniform_intensity_noise_ms1():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = UniformNoiseInjector(ppm_noise=5e-6, intensity_noise=0.4)
    scan = noise_injector.inject_noise(scan)
    assert (abs(scan.i - i_array) < i_array * 0.4).all()


# make it a test class
# integration test, test UniformNoiseInjector methods by themselves!
def test_uniform_intensity_noise_ms2():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 2,
        }
    )
    noise_injector = UniformNoiseInjector(ppm_noise=5e-6, intensity_noise=0.4)
    scan = noise_injector.inject_noise(scan, dropout=0.0)
    # breakpoint()
    print("before", i_array)
    print("after", scan.i)
    print(abs(i_array - scan.i))
    print()
    assert (abs(scan.i - i_array) < i_array * 0.4).all()


def test_uniform_intensity_noise_ms2_drop_all():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 2,
        }
    )
    # breakpoint()
    noise_injector = UniformNoiseInjector(dropout=1.0)
    scan = noise_injector.inject_noise(scan)
    # breakpoint()
    assert len(scan.mz) == 0
    assert len(scan.i) == 0


def test_UniformNoiseInjector_generate_intensity_noise():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = UniformNoiseInjector(ppm_noise=5e-6, intensity_noise=0.4)
    noise = noise_injector._generate_intensity_noise(scan, intensity_noise=0.4)
    assert (abs(noise) < i_array * 0.4).all()


def test_UniformNoiseInjector_generate_mz_noise():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = UniformNoiseInjector(ppm_noise=5e-6, intensity_noise=0.4)
    noise = noise_injector._generate_mz_noise(scan, ppm_noise=5e-6)
    # breakpoint()
    assert (abs(noise) < scan.mz * 5e-6).all()


def test_UniformNoiseInjector_mz_noise_ms1():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = UniformNoiseInjector(ppm_noise=5e-6, intensity_noise=0.4)
    scan = noise_injector.inject_noise(scan)
    # breakpoint()
    assert (
        abs(scan.mz - np.array([100, 200, 300], dtype="float32"))
        < np.array([100, 200, 300], dtype="float32") * 5e-6
    ).all()


# def test_JamssNoiseInjector_mz_noise_ms1():
#    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
#    scan = Scan(
#        {
#            "mz": np.array([100, 200, 300], dtype="float32"),
#            "i": i_array,
#            "ms_level": 1,
#        }
#    )
#    noise_injector = JamssNoiseInjector(ppm_noise=5e-6, intensity_noise=0.4)
#    scan = noise_injector.inject_noise(scan)


def test_JamssNoiseInjector_intensity_noise_ms1():
    i_array = np.array([1e6, 2e6, 3e6], dtype="float32")
    scan = Scan(
        {
            "mz": np.array([100, 200, 300], dtype="float32"),
            "i": i_array,
            "ms_level": 1,
        }
    )
    noise_injector = JamssNoiseInjector(ppm_noise=5e-6, intensity_noise=0.4)
    scan = noise_injector.inject_noise(scan)
